import ast


def set_config(engine, cmdline):
    """
    Set one or more Plover config options upon executing a stroke pattern.

    Syntax:
    {PLOVER:SET_CONFIG:option:value}
    {PLOVER:SET_CONFIG:option1:value1,option2:value2,...}

    Example usage:
    "O*EP":   "{PLOVER:SET_CONFIG:'translation_frame_opacity':100}",
    "STA*RT": "{PLOVER:SET_CONFIG:'start_attached':True,'start_capitalized':True}",

    Be careful with nested quotes. Plover's JSON dictionaries use double quotes
    by default, so use single quotes for config option names and other strings.
    """
    # Each config setting can be processed as a key:value pair in a dict.
    # The engine.config property setter will update all settings at once.
    engine.config = _cmdline_to_dict(cmdline)


def _cmdline_to_dict(cmdline):
    """Add braces and parse the entire command line as a Python dict literal."""
    try:
        opt_dict = ast.literal_eval("{" + cmdline + "}")
        assert isinstance(opt_dict, dict)
        return opt_dict
    except (AssertionError, SyntaxError, ValueError) as e:
        raise ValueError(
            'Bad command string "%s" for PLOVER:SET_CONFIG.\n' % cmdline
            + "See for reference:\n\n"
            + set_config.__doc__
        ) from e



from collections import namedtuple
from functools import total_ordering

from packaging.version import parse


@total_ordering
class PluginMetadata(
    namedtuple(
        "PluginMetadata",
        """
                                author
                                author_email
                                description
                                description_content_type
                                home_page
                                keywords
                                license
                                name
                                summary
                                version
                                """,
    )
):
    @property
    def requirement(self):
        return "%s==%s" % (self.name, self.version)

    @property
    def parsed_version(self):
        return parse(self.version)

    @classmethod
    def from_dict(cls, d):
        return cls(*(d.get(k, "") for k in cls._fields))

    @classmethod
    def from_kwargs(cls, **kwargs):
        return cls.from_dict(kwargs)

    def to_dict(self):
        return dict(zip(self._fields, self))

    def __eq__(self, other):
        return (self.name.lower(), self.parsed_version) == (
            other.name.lower(),
            other.parsed_version,
        )

    def __lt__(self, other):
        return (self.name.lower(), self.parsed_version) < (
            other.name.lower(),
            other.parsed_version,
        )



# Copyright (c) 2013 Hesky Fisher
# See LICENSE.txt for details.

"Thread-based monitoring of a stenotype machine using the passport protocol."

from itertools import zip_longest

from plover.machine.base import SerialStenotypeBase

# Passport protocol is documented here:
# http://www.eclipsecat.com/?q=system/files/Passport%20protocol_0.pdf


class Passport(SerialStenotypeBase):
    """Passport interface."""

    KEYS_LAYOUT = """
        # # # # # # # # # #
        S T P H ~ F N L Y D
        C K W R * Q B G X Z
            A O   E U
        ! ^ +
    """

    SERIAL_PARAMS = dict(SerialStenotypeBase.SERIAL_PARAMS)
    SERIAL_PARAMS.update(baudrate=38400)

    def __init__(self, params):
        super().__init__(params)
        self.packet = []

    def _read(self, b):
        b = chr(b)
        self.packet.append(b)
        if b == ">":
            self._handle_packet("".join(self.packet))
            del self.packet[:]

    def _handle_packet(self, packet):
        encoded = packet.split("/")[1]
        steno_keys = []
        for key, shadow in grouper(encoded, 2, 0):
            shadow = int(shadow, base=16)
            if shadow >= 8:
                steno_keys.append(key)
        self._notify_keys(steno_keys)

    def run(self):
        """Overrides base class run method. Do not call directly."""
        self._ready()

        while not self.finished.is_set():
            # Grab data from the serial port.
            raw = self.serial_port.read(max(1, self.serial_port.inWaiting()))

            for b in raw:
                self._read(b)


def grouper(iterable, n, fillvalue=None):
    "Collect data into fixed-length chunks or blocks"
    # grouper('ABCDEFG', 3, 'x') --> ABC DEF Gxx
    args = [iter(iterable)] * n
    return zip_longest(fillvalue=fillvalue, *args)



# Copyright (c) 2011 Hesky Fisher
# See LICENSE.txt for details.

"""Unit tests for passport.py."""

import threading

from plover.machine.passport import Passport

from plover_build_utils.testing import parametrize


class MockSerial:
    def __init__(self, **params):
        pass

    def isOpen(self):
        return True

    def inWaiting(self):
        return len(self.data)

    def read(self, size=1):
        assert size >= self.inWaiting()
        if not self.data:
            return ""
        data = self.data.pop(0)
        if not self.data:
            self.event.set()
        return data

    def close(self):
        pass


PASSPORT_TESTS = (
    # Test all keys
    lambda: (
        (b"!f#f+f*fAfCfBfEfDfGfFfHfKfLfOfNfQfPfSfRfUfTfWfYfXfZf^f~f",),
        [
            set(Passport.get_keys()),
        ],
    ),
    # Anything below 8 is not registered
    lambda: (
        (b"S9T8A7",),
        [
            ["S", "T"],
        ],
    ),
    # Sequence of strokes
    lambda: (
        (b"SfTf", b"Zf", b"QfLf"),
        [
            ["S", "T"],
            [
                "Z",
            ],
            ["Q", "L"],
        ],
    ),
)


@parametrize(PASSPORT_TESTS)
def test_passport(monkeypatch, inputs, expected):
    params = {k: v[0] for k, v in Passport.get_option_info().items()}

    class mock(MockSerial):
        event = threading.Event()
        data = [b"<123/" + s + b"/something>" for s in inputs]

    monkeypatch.setattr("plover.machine.base.serial.Serial", mock)
    actual = []
    m = Passport(params)
    m.add_stroke_callback(actual.append)
    m.start_capture()
    mock.event.wait()
    m.stop_capture()
    assert len(actual) == len(expected)
    for actual_keys, expected_keys in zip(actual, expected):
        assert sorted(actual_keys) == sorted(expected_keys)



from PySide6.QtCore import QSettings
from PySide6.QtGui import (
    QAction,
    QGuiApplication,
    QKeySequence,
)
from PySide6.QtWidgets import (
    QMainWindow,
    QToolBar,
    QToolButton,
    QWidget,
    QApplication,
)
import importlib.resources

from plover import _


def is_wayland():
    return "wayland" in QApplication.platformName().lower()


def ActionCopyViewSelectionToClipboard(view):
    def copy_selection_to_clipboard():
        indexes = view.selectedIndexes()
        data = view.model().mimeData(indexes)
        QGuiApplication.clipboard().setMimeData(data)

    action = QAction(_("Copy selection to clipboard"))
    action.setShortcut(QKeySequence(QKeySequence.StandardKey.Copy))
    action.triggered.connect(copy_selection_to_clipboard)
    return action


def ToolButton(action):
    button = QToolButton()
    button.setDefaultAction(action)
    return button


def ToolBar(*action_list):
    toolbar = QToolBar()
    for action in action_list:
        if action is None:
            toolbar.addSeparator()
        else:
            toolbar.addWidget(ToolButton(action))
    return toolbar


class WindowStateMixin:
    """
    Mixin class for saving and restoring window state using QSettings.

    This class is used as a mixin alongside a class that inherits from QWidget.
    It does NOT inherit from QWidget to avoid multiple inheritance issues.

    Usage:
        class MyDialog(QDialog, WindowStateMixin):
            ...
    """

    ROLE = None

    def _save_state(self, settings):
        """
        To be overwritten by subclasses to save additional state.
        """
        pass

    def save_state(self):
        assert self.ROLE
        assert isinstance(self, QWidget), (
            "WindowStateMixin must be used with a QWidget subclass"
        )

        settings = QSettings()
        settings.beginGroup(self.ROLE)
        settings.setValue("geometry", self.saveGeometry())
        if isinstance(self, QMainWindow):
            settings.setValue("state", self.saveState())
        self._save_state(settings)
        settings.endGroup()

    def _restore_state(self, settings):
        """
        To be overwritten by subclasses to restore additional state.
        """
        pass

    def restore_state(self):
        assert self.ROLE
        assert isinstance(self, QWidget), (
            "WindowStateMixin must be used with a QWidget subclass"
        )

        settings = QSettings()
        settings.beginGroup(self.ROLE)
        geometry = settings.value("geometry")
        if geometry is not None:
            self.restoreGeometry(geometry)
        if isinstance(self, QMainWindow):
            state = settings.value("state")
            if state is not None:
                self.restoreState(state)
        self._restore_state(settings)
        settings.endGroup()


def find_menu_actions(menu):
    """Recursively find and return a menu actions."""
    actions_dict = {}
    for action in menu.actions():
        name = action.objectName()
        if not name:
            sub_menu = action.menu()
            if sub_menu is None:
                continue
            actions_dict.update(find_menu_actions(sub_menu))
            name = sub_menu.objectName()
        assert name
        assert name not in actions_dict
        actions_dict[name] = action
    return actions_dict



from threading import Thread, current_thread

from plover.engine import StenoEngine

from plover.gui_none.add_translation import AddTranslation


class Engine(StenoEngine, Thread):
    def __init__(self, config, controller, keyboard_emulation):
        StenoEngine.__init__(self, config, controller, keyboard_emulation)
        Thread.__init__(self)
        self.name += "-engine"
        self._add_translation = AddTranslation(self)
        # self.hook_connect('quit', self.quit)

    def _in_engine_thread(self):
        return current_thread() == self

    def start(self):
        Thread.start(self)
        StenoEngine.start(self)

    def join(self):
        Thread.join(self)
        return self.code



import inspect
import itertools

import pytest

from plover.key_combo import parse_key_combo


def generate_combo_tests(test_id, *params, key_name_to_key_code=None):
    yield ("id", test_id)
    if key_name_to_key_code is not None:
        yield ("key_name_to_key_code", key_name_to_key_code)
    for iterables in params:
        iterables = [i if isinstance(i, (tuple, list)) else (i,) for i in iterables]
        if len(iterables) < 2:
            iterables.append(("",))
        for (
            combo_string,
            parse_result,
        ) in itertools.product(*iterables):
            yield (("parse", combo_string, parse_result),)


# Test directives:
# - id TEST_IDENTIFIER
# - key_name_to_key_code KEY_NAME_TO_KEY_CODE_FUNCTION
# - parse COMBO_STRING KEY_EVENTS
KEY_COMBO_TESTS = tuple(
    itertools.chain(
        (("key_name_to_key_code", lambda k: k),),
        # No-op.
        generate_combo_tests(
            "no-op",
            (("", "   "), ""),
        ),
        # Syntax error:
        generate_combo_tests(
            "syntax_error",
            (
                (
                    # - invalid character
                    "Return,",
                    "Return&",
                    "Ret. urn <",
                    "exclam ! foo",
                    "shift[a]",
                    # - unbalanced `)`
                    ") arg",
                    "arg )",
                    "arg())",
                    "arg(x) )",
                    # - unbalanced `(`
                    "test(",
                    "( grr",
                    "foo ( bar",
                    "foo (bar ( ",
                    "foo ((",
                    # - [-+]key() is not valid
                    "+foo ()",
                    "+foo()",
                ),
                SyntaxError,
            ),
        ),
        # Pressing an already pressed key.
        generate_combo_tests(
            "already_pressed",
            (
                (
                    "foo(foo)",
                    "Foo(foO)",
                    "foo(fOo(arg))",
                    "foo(bar(Foo))",
                    "foo(bar(foo(x)))",
                ),
                ValueError,
            ),
        ),
        # Stacking.
        generate_combo_tests(
            "stacking",
            # 1 is not a valid identifier, but still a valid key name.
            ("1", "+1 -1"),
            (("Shift_l", "SHIFT_L"), "+shift_l -shift_l"),
            # Case does not matter.
            (("a", " A "), "+a -a"),
            (("a(b c)", "a ( b c   )"), "+a +b -b +c -c -a"),
            (("a(bc)", " a(  Bc )"), "+a +bc -bc -a"),
            (
                ("a(bc(d)e f(g) h())i j"),
                "+a +bc +d -d -bc +e -e +f +g -g -f +h -h -a +i -i +j -j",
            ),
            (
                ("foo () bar ( foo a b c (d))", "fOo () Bar ( FOO a B c (D))"),
                "+foo -foo +bar +foo -foo +a -a +b -b +c +d -d -c -bar",
            ),
        ),
        # Invalid key name.
        generate_combo_tests(
            "invalid_key",
            ("1 (c) 2 bad 3 (a b c)", ValueError),
            key_name_to_key_code={c: c for c in "123abc"}.get,
        ),
        # Same key code, multiple key names.
        generate_combo_tests(
            "aliasing",
            ("1 exclam", "+10 -10 +10 -10"),
            (("1 ( exclam )", "exclam(1)"), ValueError),
            key_name_to_key_code={"1": "10", "exclam": "10"}.get,
        ),
    )
)


def parametrize(tests):
    key_name_to_key_code = None
    test_id = None
    args = []
    ids = []
    for t in tests:
        if t[0] == "key_name_to_key_code":
            key_name_to_key_code = t[1]
        elif t[0] == "id":
            test_id = t[1]
        else:
            assert key_name_to_key_code is not None
            assert test_id is not None
            args.append((key_name_to_key_code, t))
            ids.append(test_id)
    return pytest.mark.parametrize(
        ("key_name_to_key_code", "instructions"), args, ids=ids
    )


@parametrize(KEY_COMBO_TESTS)
def test_key_combo(key_name_to_key_code, instructions):
    def repr_expected(result):
        assert isinstance(result, str)
        return [s.strip() for s in result.split()]

    def repr_key_events(events):
        assert isinstance(events, list)
        return ["%s%s" % ("+" if pressed else "-", key) for key, pressed in events]

    for action, *args in instructions:
        if action == "parse":
            combo_string, key_events = args
            if inspect.isclass(key_events):
                with pytest.raises(key_events):
                    parse_key_combo(
                        combo_string, key_name_to_key_code=key_name_to_key_code
                    )
            else:
                assert repr_key_events(
                    parse_key_combo(
                        combo_string, key_name_to_key_code=key_name_to_key_code
                    )
                ) == repr_expected(key_events)
        else:
            raise ValueError(args[0])



from .display_server import DISPLAY_SERVER

if DISPLAY_SERVER == "x11":
    from .keyboardcontrol_x11 import KeyboardCapture, KeyboardEmulation  # pylint: disable=unused-import
else:
    from .keyboardcontrol_uinput import KeyboardCapture, KeyboardEmulation  # pylint: disable=unused-import



from evdev import (
    UInput,
    ecodes as e,
    util,
    InputDevice,
    list_devices,
    InputEvent,
    KeyEvent,
)
import threading
import os
import selectors

from psutil import process_iter

from plover.output.keyboard import GenericKeyboardEmulation
from plover.machine.keyboard_capture import Capture
from plover.key_combo import parse_key_combo
from plover import log

# Shared keys between all layouts
BASE_LAYOUT = {
    # Modifiers
    "alt_l": e.KEY_LEFTALT,
    "alt_r": e.KEY_RIGHTALT,
    "alt": e.KEY_LEFTALT,
    "ctrl_l": e.KEY_LEFTCTRL,
    "ctrl_r": e.KEY_RIGHTCTRL,
    "ctrl": e.KEY_LEFTCTRL,
    "control_l": e.KEY_LEFTCTRL,
    "control_r": e.KEY_RIGHTCTRL,
    "control": e.KEY_LEFTCTRL,
    "shift_l": e.KEY_LEFTSHIFT,
    "shift_r": e.KEY_RIGHTSHIFT,
    "shift": e.KEY_LEFTSHIFT,
    "super_l": e.KEY_LEFTMETA,
    "super_r": e.KEY_RIGHTMETA,
    "super": e.KEY_LEFTMETA,
    # Numbers
    "1": e.KEY_1,
    "2": e.KEY_2,
    "3": e.KEY_3,
    "4": e.KEY_4,
    "5": e.KEY_5,
    "6": e.KEY_6,
    "7": e.KEY_7,
    "8": e.KEY_8,
    "9": e.KEY_9,
    "0": e.KEY_0,
    # Symbols
    " ": e.KEY_SPACE,
    "\b": e.KEY_BACKSPACE,
    "\n": e.KEY_ENTER,
    # https://github.com/openstenoproject/plover/blob/9b5a357f1fb57cb0a9a8596ae12cd1e84fcff6c4/plover/oslayer/osx/keyboardcontrol.py#L75
    # https://gist.github.com/jfortin42/68a1fcbf7738a1819eb4b2eef298f4f8
    "return": e.KEY_ENTER,
    "tab": e.KEY_TAB,
    "backspace": e.KEY_BACKSPACE,
    "delete": e.KEY_DELETE,
    "escape": e.KEY_ESC,
    "clear": e.KEY_CLEAR,
    "minus": e.KEY_MINUS,
    "equal": e.KEY_EQUAL,
    "bracketleft": e.KEY_LEFTBRACE,
    "bracketright": e.KEY_RIGHTBRACE,
    "backslash": e.KEY_BACKSLASH,
    "semicolon": e.KEY_SEMICOLON,
    "apostrophe": e.KEY_APOSTROPHE,
    "comma": e.KEY_COMMA,
    "dot": e.KEY_DOT,
    "slash": e.KEY_SLASH,
    "grave": e.KEY_GRAVE,
    "-": e.KEY_MINUS,
    "=": e.KEY_EQUAL,
    "[": e.KEY_LEFTBRACE,
    "]": e.KEY_RIGHTBRACE,
    "\\": e.KEY_BACKSLASH,
    ";": e.KEY_SEMICOLON,
    "'": e.KEY_APOSTROPHE,
    ",": e.KEY_COMMA,
    ".": e.KEY_DOT,
    "/": e.KEY_SLASH,
    "`": e.KEY_GRAVE,
    # Navigation
    "up": e.KEY_UP,
    "down": e.KEY_DOWN,
    "left": e.KEY_LEFT,
    "right": e.KEY_RIGHT,
    "page_up": e.KEY_PAGEUP,
    "page_down": e.KEY_PAGEDOWN,
    "home": e.KEY_HOME,
    "insert": e.KEY_INSERT,
    "end": e.KEY_END,
    "space": e.KEY_SPACE,
    "print": e.KEY_PRINT,
    # Function keys
    "fn": e.KEY_FN,
    "f1": e.KEY_F1,
    "f2": e.KEY_F2,
    "f3": e.KEY_F3,
    "f4": e.KEY_F4,
    "f5": e.KEY_F5,
    "f6": e.KEY_F6,
    "f7": e.KEY_F7,
    "f8": e.KEY_F8,
    "f9": e.KEY_F9,
    "f10": e.KEY_F10,
    "f11": e.KEY_F11,
    "f12": e.KEY_F12,
    "f13": e.KEY_F13,
    "f14": e.KEY_F14,
    "f15": e.KEY_F15,
    "f16": e.KEY_F16,
    "f17": e.KEY_F17,
    "f18": e.KEY_F18,
    "f19": e.KEY_F19,
    "f20": e.KEY_F20,
    "f21": e.KEY_F21,
    "f22": e.KEY_F22,
    "f23": e.KEY_F23,
    "f24": e.KEY_F24,
    # Numpad
    "kp_1": e.KEY_KP1,
    "kp_2": e.KEY_KP2,
    "kp_3": e.KEY_KP3,
    "kp_4": e.KEY_KP4,
    "kp_5": e.KEY_KP5,
    "kp_6": e.KEY_KP6,
    "kp_7": e.KEY_KP7,
    "kp_8": e.KEY_KP8,
    "kp_9": e.KEY_KP9,
    "kp_0": e.KEY_KP0,
    "kp_add": e.KEY_KPPLUS,
    "kp_decimal": e.KEY_KPDOT,
    "kp_delete": e.KEY_DELETE,  # There is no KPDELETE
    "kp_divide": e.KEY_KPSLASH,
    "kp_enter": e.KEY_KPENTER,
    "kp_equal": e.KEY_KPEQUAL,
    "kp_multiply": e.KEY_KPASTERISK,
    "kp_subtract": e.KEY_KPMINUS,
    # Media keys
    "audioraisevolume": e.KEY_VOLUMEUP,
    "audiolowervolume": e.KEY_VOLUMEDOWN,
    "monbrightnessup": e.KEY_BRIGHTNESSUP,
    "monbrightnessdown": e.KEY_BRIGHTNESSDOWN,
    "audiomute": e.KEY_MUTE,
    "num_lock": e.KEY_NUMLOCK,
    "eject": e.KEY_EJECTCD,
    "audiopause": e.KEY_PAUSE,
    "audioplay": e.KEY_PLAY,
    "audionext": e.KEY_NEXT,
    "audiorewind": e.KEY_REWIND,
    "kbdbrightnessup": e.KEY_KBDILLUMUP,
    "kbdbrightnessdown": e.KEY_KBDILLUMDOWN,
}

DEFAULT_LAYOUT = "qwerty"
LAYOUTS = {
    # Only specify keys that differ from qwerty
    "qwerty": {
        **BASE_LAYOUT,
        # Top row
        "q": e.KEY_Q,
        "w": e.KEY_W,
        "e": e.KEY_E,
        "r": e.KEY_R,
        "t": e.KEY_T,
        "y": e.KEY_Y,
        "u": e.KEY_U,
        "i": e.KEY_I,
        "o": e.KEY_O,
        "p": e.KEY_P,
        # Middle row
        "a": e.KEY_A,
        "s": e.KEY_S,
        "d": e.KEY_D,
        "f": e.KEY_F,
        "g": e.KEY_G,
        "h": e.KEY_H,
        "j": e.KEY_J,
        "k": e.KEY_K,
        "l": e.KEY_L,
        # Bottom row
        "z": e.KEY_Z,
        "x": e.KEY_X,
        "c": e.KEY_C,
        "v": e.KEY_V,
        "b": e.KEY_B,
        "n": e.KEY_N,
        "m": e.KEY_M,
    },
    "qwertz": {
        **BASE_LAYOUT,
        # Top row
        "q": e.KEY_Q,
        "w": e.KEY_W,
        "e": e.KEY_E,
        "r": e.KEY_R,
        "t": e.KEY_T,
        "z": e.KEY_Y,
        "u": e.KEY_U,
        "i": e.KEY_I,
        "o": e.KEY_O,
        "p": e.KEY_P,
        # Middle row
        "a": e.KEY_A,
        "s": e.KEY_S,
        "d": e.KEY_D,
        "f": e.KEY_F,
        "g": e.KEY_G,
        "h": e.KEY_H,
        "j": e.KEY_J,
        "k": e.KEY_K,
        "l": e.KEY_L,
        # Bottom row
        "y": e.KEY_Z,
        "x": e.KEY_X,
        "c": e.KEY_C,
        "v": e.KEY_V,
        "b": e.KEY_B,
        "n": e.KEY_N,
        "m": e.KEY_M,
    },
    "colemak": {
        **BASE_LAYOUT,
        # Top row
        "q": e.KEY_Q,
        "w": e.KEY_W,
        "f": e.KEY_E,
        "p": e.KEY_R,
        "g": e.KEY_T,
        "j": e.KEY_Y,
        "l": e.KEY_U,
        "u": e.KEY_I,
        "y": e.KEY_O,
        # Middle row
        "a": e.KEY_A,
        "r": e.KEY_S,
        "s": e.KEY_D,
        "t": e.KEY_F,
        "d": e.KEY_G,
        "h": e.KEY_H,
        "n": e.KEY_J,
        "e": e.KEY_K,
        "i": e.KEY_L,
        "o": e.KEY_SEMICOLON,
        # Bottom row
        "z": e.KEY_Z,
        "x": e.KEY_X,
        "c": e.KEY_C,
        "v": e.KEY_V,
        "b": e.KEY_B,
        "k": e.KEY_N,
        "m": e.KEY_M,
    },
    "colemak-dh": {
        **BASE_LAYOUT,
        # Top row
        "q": e.KEY_Q,
        "w": e.KEY_W,
        "f": e.KEY_E,
        "p": e.KEY_R,
        "b": e.KEY_T,
        "j": e.KEY_Y,
        "l": e.KEY_U,
        "u": e.KEY_I,
        "y": e.KEY_O,
        # Middle row
        "a": e.KEY_A,
        "r": e.KEY_S,
        "s": e.KEY_D,
        "t": e.KEY_F,
        "g": e.KEY_G,
        "m": e.KEY_H,
        "n": e.KEY_J,
        "e": e.KEY_K,
        "i": e.KEY_L,
        "o": e.KEY_SEMICOLON,
        # Bottom row
        "z": e.KEY_BACKSLASH,  # less than-key
        "x": e.KEY_Z,
        "c": e.KEY_X,
        "d": e.KEY_C,
        "v": e.KEY_V,
        "k": e.KEY_N,
        "h": e.KEY_M,
    },
    "dvorak": {
        **BASE_LAYOUT,
        # Top row
        "'": e.KEY_Q,
        ",": e.KEY_W,
        ".": e.KEY_E,
        "p": e.KEY_R,
        "y": e.KEY_T,
        "f": e.KEY_Y,
        "g": e.KEY_U,
        "c": e.KEY_I,
        "r": e.KEY_O,
        "l": e.KEY_P,
        "/": e.KEY_LEFTBRACE,
        "=": e.KEY_RIGHTBRACE,
        # Middle row
        "a": e.KEY_A,
        "o": e.KEY_S,
        "e": e.KEY_D,
        "u": e.KEY_F,
        "i": e.KEY_G,
        "d": e.KEY_H,
        "h": e.KEY_J,
        "t": e.KEY_K,
        "n": e.KEY_L,
        "s": e.KEY_SEMICOLON,
        "-": e.KEY_APOSTROPHE,
        # Bottom row
        ";": e.KEY_Z,
        "q": e.KEY_X,
        "j": e.KEY_C,
        "k": e.KEY_V,
        "x": e.KEY_B,
        "b": e.KEY_N,
        "m": e.KEY_M,
        "w": e.KEY_COMMA,
        "v": e.KEY_DOT,
        "z": e.KEY_SLASH,
    },
}

KEYCODE_TO_KEY = dict(
    zip(LAYOUTS[DEFAULT_LAYOUT].values(), LAYOUTS[DEFAULT_LAYOUT].keys())
)

MODIFIER_KEY_CODES: set[int] = {
    e.KEY_LEFTSHIFT,
    e.KEY_RIGHTSHIFT,
    e.KEY_LEFTCTRL,
    e.KEY_RIGHTCTRL,
    e.KEY_LEFTALT,
    e.KEY_RIGHTALT,
    e.KEY_LEFTMETA,
    e.KEY_RIGHTMETA,
}


class KeyboardEmulation(GenericKeyboardEmulation):
    def __init__(self):
        super().__init__()
        # Initialize UInput with all keys available
        self._res = util.find_ecodes_by_regex(r"KEY_.*")
        self._ui = UInput(self._res)

        # Check that ibus or fcitx5 is running
        if not any(p.name() in ["ibus-daemon", "fcitx5"] for p in process_iter()):
            log.warning(
                "It appears that an input method, such as ibus or fcitx5, is not running on your system. Without this, some text may not be output correctly."
            )

    def _update_layout(self, layout):
        if not layout in LAYOUTS:
            log.warning(f"Layout {layout} not supported. Falling back to qwerty.")
        self._KEY_TO_KEYCODE = LAYOUTS.get(layout, LAYOUTS[DEFAULT_LAYOUT])

    def _get_key(self, key):
        """Helper function to get the keycode and potential shift key for uppercase."""
        if key in self._KEY_TO_KEYCODE:
            return (self._KEY_TO_KEYCODE[key], [])
        elif key.lower() in self._KEY_TO_KEYCODE:
            # mods is a list for the potential of expanding it in the future to include altgr
            return (
                self._KEY_TO_KEYCODE[key.lower()],
                [self._KEY_TO_KEYCODE["shift_l"]],
            )
        return (None, [])

    def _press_key(self, key, state):
        self._ui.write(e.EV_KEY, key, 1 if state else 0)
        self._ui.syn()

    """
    Send a unicode character.
    This depends on an IME such as iBus or fcitx5. iBus is used by GNOME, and fcitx5 by KDE.
    It assumes the default keybinding ctrl-shift-u, enter hex, enter is used, which is the default in both.
    From my testing, it works fine in using iBus and fcitx5, but in kitty terminal emulator, which uses
    the same keybinding, it's too fast for it to handle and ends up writing random stuff. I don't
    think there is a way to fix that other than increasing the delay.
    """

    def _send_unicode(self, hex):
        self.send_key_combination("ctrl_l(shift(u))")
        self.delay()
        self.send_string(hex)
        self.delay()
        self._send_char(" ")

    def _send_char(self, char):
        (base, mods) = self._get_key(char)

        # Key can be sent with a key combination
        if base is not None:
            for mod in mods:
                self._press_key(mod, True)
            self.delay()
            self._press_key(base, True)
            self._press_key(base, False)
            for mod in mods:
                self._press_key(mod, False)

        # Key press can not be emulated - send unicode symbol instead
        else:
            # Convert to hex and remove leading "0x"
            unicode_hex = hex(ord(char))[2:]
            self._send_unicode(unicode_hex)

    def send_string(self, string):
        for key in self.with_delay(list(string)):
            self._send_char(key)

    def send_backspaces(self, count):
        for _ in range(count):
            self._send_char("\b")

    def send_key_combination(self, combo):
        # https://plover.readthedocs.io/en/latest/api/key_combo.html#module-plover.key_combo
        key_events = parse_key_combo(combo)

        for key, pressed in self.with_delay(key_events):
            (base, _) = self._get_key(key)

            if base is not None:
                self._press_key(base, pressed)
            else:
                log.warning("Key " + key + " is not valid!")


class KeyboardCapture(Capture):
    _selector: selectors.DefaultSelector
    _device_thread: threading.Thread | None
    # Pipes to signal `_run` thread to stop
    _device_thread_read_pipe: int | None
    _device_thread_write_pipe: int | None

    def __init__(self):
        super().__init__()
        self._devices = self._get_devices()

        self._selector = selectors.DefaultSelector()
        self._device_thread = None
        self._device_thread_read_pipe = None
        self._device_thread_write_pipe = None

        self._res = util.find_ecodes_by_regex(r"KEY_.*")
        self._ui = UInput(self._res)
        self._suppressed_keys = set()
        # The keycodes from evdev, e.g. e.KEY_A refers to the *physical* a, which corresponds with the qwerty layout.

    def _get_devices(self):
        input_devices = [InputDevice(path) for path in list_devices()]
        keyboard_devices = [dev for dev in input_devices if self._filter_devices(dev)]
        return keyboard_devices

    def _filter_devices(self, device):
        """
        Filter out devices that should not be grabbed and suppressed, to avoid output feeding into itself.
        """
        is_uinput = device.name == "py-evdev-uinput" or device.phys == "py-evdev-uinput"
        # Check for some common keys to make sure it's really a keyboard
        keys = device.capabilities().get(e.EV_KEY, [])
        keyboard_keys_present = any(
            key in keys
            for key in [e.KEY_ESC, e.KEY_SPACE, e.KEY_ENTER, e.KEY_LEFTSHIFT]
        )
        return not is_uinput and keyboard_keys_present

    def _grab_devices(self):
        """Grab all devices, waiting for each device to stop having keys pressed.

        If a device is grabbed when keys are being pressed, the key will
        appear to be always pressed down until the device is ungrabbed and the
        key is pressed again.
        See https://stackoverflow.com/questions/41995349/why-does-ioctlfd-eviocgrab-1-cause-key-spam-sometimes
        There is likely a race condition here between checking active keys and
        actually grabbing the device, but it appears to work fine.
        """
        for device in self._devices:
            if len(device.active_keys()) > 0:
                for _ in device.read_loop():
                    if len(device.active_keys()) == 0:
                        # No keys are pressed. Grab the device
                        break
            device.grab()

    def _ungrab_devices(self):
        """Ungrab all devices. Handles all exceptions when ungrabbing."""
        for device in self._devices:
            try:
                device.ungrab()
            except:
                log.debug("failed to ungrab device", exc_info=True)

    def start(self):
        # Exception handling note: cancel() will eventually be called when the
        # machine reconnect button is pressed or when the machine is changed.
        # Therefore, cancel() does not need to be called in the except block.
        try:
            self._grab_devices()
            self._device_thread_read_pipe, self._device_thread_write_pipe = os.pipe()
            self._selector.register(self._device_thread_read_pipe, selectors.EVENT_READ)
            for device in self._devices:
                self._selector.register(device, selectors.EVENT_READ)

            self._device_thread = threading.Thread(target=self._run)
            self._device_thread.start()
        except Exception:
            self._ungrab_devices()
            self._ui.close()
            raise

    def cancel(self):
        if (
            self._device_thread_read_pipe is None
            or self._device_thread_write_pipe is None
        ):
            # The only way for these pipes to be None is if pipe creation in start() failed
            # In that case, no other code after pipe creation would have run
            # and no cleanup is required
            return
        try:
            # Write some arbitrary data to the pipe to signal the _run thread to stop
            os.write(self._device_thread_write_pipe, b"a")
            if self._device_thread is not None:
                self._device_thread.join()
            self._selector.close()
        except Exception:
            log.debug("error stopping KeyboardCapture", exc_info=True)
        finally:
            os.close(self._device_thread_read_pipe)
            os.close(self._device_thread_write_pipe)

    def suppress(self, suppressed_keys=()):
        """
        UInput is not capable of suppressing only specific keys. To get around this, non-suppressed keys
        are passed through to a UInput device and emulated, while keys in this list get sent to plover.
        It does add a little bit of delay, but that is not noticeable.
        """
        self._suppressed_keys = set(suppressed_keys)

    def _run(self):
        keys_pressed_with_modifier: set[int] = set()
        down_modifier_keys: set[int] = set()

        def _process_key_event(event: InputEvent) -> tuple[str | None, bool]:
            """
            Processes an InputEvent to determine which key Plover should receive
            and whether the event should be suppressed.
            Considers pressed modifiers and Plover's suppressed keys.
            Returns a tuple of (key_to_send_to_plover, suppress)
            """
            if not self._suppressed_keys:
                # No keys are suppressed
                # Always send to Plover so that it can handle global shortcuts like PLOVER_TOGGLE (PHROLG)
                return KEYCODE_TO_KEY.get(event.code, None), False
            if event.code in MODIFIER_KEY_CODES:
                # Can't use if-else because there is a third case: key_hold
                if event.value == KeyEvent.key_down:
                    down_modifier_keys.add(event.code)
                elif event.value == KeyEvent.key_up:
                    down_modifier_keys.discard(event.code)
                return None, False
            key = KEYCODE_TO_KEY.get(event.code, None)
            if key is None:
                # Key is unhandled. Passthrough
                return None, False
            if event.value == KeyEvent.key_down and down_modifier_keys:
                keys_pressed_with_modifier.add(event.code)
                return None, False
            if (
                event.value == KeyEvent.key_up
                and event.code in keys_pressed_with_modifier
            ):
                # Must pass through key up event if key was pressed with modifier
                # or else it will stay pressed down and start repeating.
                # Must release even if modifier key was released first
                keys_pressed_with_modifier.discard(event.code)
                return None, False
            suppress = key in self._suppressed_keys
            return key, suppress

        try:
            while True:
                for key, events in self._selector.select():
                    if key.fd == self._device_thread_read_pipe:
                        # Stop this thread
                        return
                    assert isinstance(key.fileobj, InputDevice)
                    device: InputDevice = key.fileobj
                    for event in device.read():
                        if event.type == e.EV_KEY:
                            key_to_send_to_plover, suppress = _process_key_event(event)
                            if key_to_send_to_plover is not None:
                                # Always send keys to Plover when no keys suppressed.
                                # This is required for global shortcuts like
                                # Plover toggle (PHROLG) when Plover is disabled.
                                # Note: Must explicitly check key_up or key_down
                                # because there is a third case: key_hold
                                if event.value == KeyEvent.key_down:
                                    self.key_down(key_to_send_to_plover)
                                elif event.value == KeyEvent.key_up:
                                    self.key_up(key_to_send_to_plover)
                            if suppress:
                                # Skip rest of loop to prevent event from
                                # being passed through
                                continue

                        # Passthrough event
                        self._ui.write_event(event)
        except:
            log.error("keyboard capture error", exc_info=True)
        finally:
            # Always ungrab devices to prevent exceptions in the _run loop
            # from causing grabbed input devices to be blocked
            self._ungrab_devices()
            self._ui.close()



import re


def gettext(contents):
    # replace ``_translate("context", `` by ``_(``
    contents = re.sub(
        r"\n", ('\n_ = __import__(__package__.split(".", 1)[0])._\n'), contents, 1
    )
    contents = re.sub(
        r"\n\s+_translate = QtCore\.QCoreApplication\.translate\n", "\n", contents
    )

    def repl(m):
        gd = m.groupdict()
        comment = "{ws}# i18n: Widget: “{widget}”".format(**gd)
        field = gd["field"]
        if field:
            field = " ".join(
                word.lower() for word in re.split(r"([A-Z][a-z_0-9]+)", field) if word
            )
            comment += ", {field}".format(field=field)
        comment += "."
        gd["pre2"] = gd["pre2"] or ""
        return "{comment}\n{ws}{pre1}{pre2}_(".format(comment=comment, **gd)

    contents = re.sub(
        (
            r"(?P<ws> *)(?P<pre1>.*?)(?P<pre2>\.set(?P<field>\w+)\()?"
            r'\b_translate\("(?P<widget>.*)",\s'
        ),
        repl,
        contents,
    )
    assert re.search(r"\b_translate\(", contents) is None
    return contents


def no_autoconnection(contents):
    # remove calls to ``QtCore.QMetaObject.connectSlotsByName``
    contents = re.sub(
        r"\n\s+QtCore\.QMetaObject\.connectSlotsByName\(\w+\)\n", "\n", contents
    )
    return contents



from PySide6.QtCore import (
    QAbstractListModel,
    QMimeData,
    QModelIndex,
    Qt,
)
from PySide6.QtGui import (
    QFont,
    QFontMetrics,
    QTextCharFormat,
    QTextCursor,
    QTextDocument,
)
from PySide6.QtWidgets import (
    QListView,
    QStyle,
    QStyledItemDelegate,
)

from plover import _
from plover.translation import escape_translation

from .utils import ActionCopyViewSelectionToClipboard


# i18n: Widget: “SuggestionsWidget”.
NO_SUGGESTIONS_STRING = _("no suggestions")
MAX_SUGGESTIONS_COUNT = 10


class SuggestionsDelegate(QStyledItemDelegate):
    def __init__(self, parent=None):
        super().__init__(parent=parent)
        self._doc = QTextDocument()
        self._translation_char_format = QTextCharFormat()
        self._strokes_char_format = QTextCharFormat()
        self._strokes_char_format.font().setStyleHint(QFont.StyleHint.Monospace)
        self._size_hint_cache = {}

    def clear_size_hint_cache(self):
        self._size_hint_cache.clear()

    @property
    def text_font(self):
        return self._translation_char_format.font()

    @text_font.setter
    def text_font(self, font):
        self._translation_char_format.setFont(font)
        self.clear_size_hint_cache()

    @property
    def strokes_font(self):
        return self._strokes_char_format.font()

    @strokes_font.setter
    def strokes_font(self, font):
        self._strokes_char_format.setFont(font)
        self.clear_size_hint_cache()

    def _format_suggestion(self, index):
        suggestion = index.data(Qt.ItemDataRole.DisplayRole)
        translation = escape_translation(suggestion.text) + ":"
        if not suggestion.steno_list:
            translation += " " + NO_SUGGESTIONS_STRING
            return translation, None
        strokes = ""
        for strokes_list in suggestion.steno_list[:MAX_SUGGESTIONS_COUNT]:
            strokes += "\n    " + "/".join(strokes_list)
        return translation, strokes

    def _suggestion_size_hint(self, index):
        translation, strokes = self._format_suggestion(index)
        size = QFontMetrics(self.text_font).size(0, translation)
        if strokes is not None:
            size += QFontMetrics(self.strokes_font).size(0, strokes)
        return size

    def paint(self, painter, option, index):
        painter.save()
        if option.state & QStyle.StateFlag.State_Selected:
            painter.fillRect(option.rect, option.palette.highlight())
            text_color = option.palette.highlightedText()
        else:
            text_color = option.palette.text()
        self._translation_char_format.setForeground(text_color)
        self._strokes_char_format.setForeground(text_color)
        painter.translate(option.rect.topLeft())
        self._doc.clear()
        cursor = QTextCursor(self._doc)
        translation, strokes = self._format_suggestion(index)
        cursor.insertText(translation, self._translation_char_format)
        if strokes is not None:
            cursor.insertText(strokes, self._strokes_char_format)
        self._doc.drawContents(painter)
        painter.restore()

    def sizeHint(self, option, index):
        size = self._size_hint_cache.get(index.row())
        if size is not None:
            return size
        size = self._suggestion_size_hint(index)
        self._size_hint_cache[index.row()] = size
        return size


class SuggestionsModel(QAbstractListModel):
    def __init__(self):
        super().__init__()
        self._suggestion_list = []

    def rowCount(self, parent):
        return 0 if parent.isValid() else len(self._suggestion_list)

    def data(self, index, role):
        if not index.isValid():
            return None
        suggestion = self._suggestion_list[index.row()]
        if role == Qt.ItemDataRole.DisplayRole:
            return suggestion
        if role == Qt.ItemDataRole.AccessibleTextRole:
            translation = escape_translation(suggestion.text)
            if suggestion.steno_list:
                steno = ", ".join(
                    "/".join(strokes_list)
                    for strokes_list in suggestion.steno_list[:MAX_SUGGESTIONS_COUNT]
                )
            else:
                steno = NO_SUGGESTIONS_STRING
            return translation + ": " + steno
        return None

    def clear(self):
        self.modelAboutToBeReset.emit()
        self._suggestion_list.clear()
        self.modelReset.emit()

    def extend(self, suggestion_list):
        row = len(self._suggestion_list)
        self.beginInsertRows(QModelIndex(), row, row + len(suggestion_list))
        self._suggestion_list.extend(suggestion_list)
        self.endInsertRows()

    def mimeTypes(self):
        return ["text/plain"]

    def mimeData(self, indexes):
        data = QMimeData()
        data.setText(
            "\n".join(
                filter(
                    None,
                    (
                        self.data(index, Qt.ItemDataRole.AccessibleTextRole)
                        for index in indexes
                    ),
                )
            )
        )
        return data


class SuggestionsWidget(QListView):
    def __init__(self, parent=None):
        super().__init__(parent=parent)
        self.setResizeMode(self.ResizeMode.Adjust)
        self.setSelectionMode(self.SelectionMode.ExtendedSelection)
        self._copy_action = ActionCopyViewSelectionToClipboard(self)
        self.addAction(self._copy_action)
        self._model = SuggestionsModel()
        self._delegate = SuggestionsDelegate(self)
        self.setModel(self._model)
        self.setItemDelegate(self._delegate)

    def append(self, suggestion_list):
        scrollbar = self.verticalScrollBar()
        scroll_at_end = scrollbar.value() == scrollbar.maximum()
        self._model.extend(suggestion_list)
        if scroll_at_end:
            self.scrollToBottom()

    def clear(self):
        self._model.clear()
        self._delegate.clear_size_hint_cache()

    def _reformat(self):
        self._model.layoutAboutToBeChanged.emit()
        self._model.layoutChanged.emit()

    @property
    def text_font(self):
        return self._delegate.text_font

    @text_font.setter
    def text_font(self, font):
        self._delegate.text_font = font
        self._reformat()

    @property
    def strokes_font(self):
        return self._delegate.strokes_font

    @strokes_font.setter
    def strokes_font(self, font):
        self._delegate.strokes_font = font
        self._reformat()



# Copyright (c) 2013 Hesky Fisher
# See LICENSE.txt for details.

import os
from logging import Handler
from collections import defaultdict

import pytest

from plover.steno import Stroke
from plover import log


class FakeHandler(Handler):
    outputs = defaultdict(list)

    def __init__(self, filename, format=log.STROKE_LOG_FORMAT):
        super().__init__()
        self.baseFilename = filename

    def emit(self, record):
        FakeHandler.outputs[self.baseFilename].append(record.getMessage())


@pytest.fixture(autouse=True)
def fake_file_log(monkeypatch):
    monkeypatch.setattr("plover.log.FileHandler", FakeHandler)
    yield
    FakeHandler.outputs.clear()
    # Reset logger state.
    log.set_stroke_filename(None)
    log.enable_stroke_logging(False)
    log.enable_translation_logging(False)


def stroke_filename(path):
    return os.path.abspath(path)


def test_set_filename():
    sf1 = stroke_filename("/fn1")
    log.set_stroke_filename("/fn1")
    log.enable_stroke_logging(True)
    log.stroke(Stroke(("S-",)))
    sf2 = stroke_filename("/fn2")
    log.set_stroke_filename("/fn2")
    log.stroke(Stroke(("-T",)))
    log.set_stroke_filename(None)
    log.stroke(Stroke(("P-",)))
    assert FakeHandler.outputs == {
        sf1: ["Stroke(S : ['S-'])"],
        sf2: ["Stroke(-T : ['-T'])"],
    }


def test_stroke():
    sf = stroke_filename("/fn")
    log.set_stroke_filename(sf)
    log.enable_stroke_logging(True)
    log.stroke(Stroke(("S-", "-T", "T-")))
    log.stroke(Stroke(("#", "S-", "-T")))
    assert FakeHandler.outputs == {
        sf: ["Stroke(ST-T : ['S-', 'T-', '-T'])", "Stroke(1-9 : ['#', 'S-', '-T'])"],
    }


def test_log_translation():
    sf = stroke_filename("/fn")
    log.set_stroke_filename(sf)
    log.enable_translation_logging(True)
    log.translation(["a", "b"], ["c", "d"], None)
    assert FakeHandler.outputs == {sf: ["*a", "*b", "c", "d"]}


def test_enable_stroke_logging():
    sf = stroke_filename("/fn")
    log.set_stroke_filename(sf)
    log.stroke(Stroke(("S-",)))
    log.enable_stroke_logging(True)
    log.stroke(Stroke(("T-",)))
    log.enable_stroke_logging(False)
    log.stroke(Stroke(("K-",)))
    assert FakeHandler.outputs == {sf: ["Stroke(T : ['T-'])"]}


def test_enable_translation_logging():
    sf = stroke_filename("/fn")
    log.set_stroke_filename(sf)
    log.translation(["a"], ["b"], None)
    log.enable_translation_logging(True)
    log.translation(["c"], ["d"], None)
    log.enable_translation_logging(False)
    log.translation(["e"], ["f"], None)
    assert FakeHandler.outputs == {sf: ["*c", "d"]}



import time

from PySide6.QtCore import (
    QAbstractListModel,
    QMimeData,
    QModelIndex,
    Qt,
    Slot,
)
from PySide6.QtGui import QFont, QFontDatabase
from PySide6.QtWidgets import (
    QFileDialog,
    QFontDialog,
    QMessageBox,
)

from wcwidth import wcwidth

from plover import _, system
from plover.steno import Stroke

from plover.gui_qt import utils
from plover.gui_qt.paper_tape_ui import Ui_PaperTape
from plover.gui_qt.utils import ActionCopyViewSelectionToClipboard, ToolBar
from plover.gui_qt.tool import Tool


STYLE_PAPER, STYLE_RAW = (
    # i18n: Paper tape style.
    _("Paper"),
    # i18n: Paper tape style.
    _("Raw"),
)
TAPE_STYLES = (STYLE_PAPER, STYLE_RAW)


class TapeModel(QAbstractListModel):
    def __init__(self):
        super().__init__()
        self._stroke_list = []
        self._style = None
        self._all_keys = None
        self._numbers = None

    def rowCount(self, parent):
        return 0 if parent.isValid() else len(self._stroke_list)

    @property
    def style(self):
        return self._style

    @style.setter
    def style(self, style):
        assert style in TAPE_STYLES
        self.layoutAboutToBeChanged.emit()
        self._style = style
        self.layoutChanged.emit()

    def _paper_format(self, stroke):
        text = self._all_keys_filler * 1
        keys = stroke.steno_keys[:]
        if any(key in self._numbers for key in keys):
            keys.append("#")
        for key in keys:
            index = system.KEY_ORDER[key]
            text[index] = self._all_keys[index]
        return "".join(text)

    @staticmethod
    def _raw_format(stroke):
        return stroke.rtfcre

    def data(self, index, role):
        if not index.isValid():
            return None
        stroke = self._stroke_list[index.row()]
        if role == Qt.ItemDataRole.DisplayRole:
            if self._style == STYLE_PAPER:
                return self._paper_format(stroke)
            if self._style == STYLE_RAW:
                return self._raw_format(stroke)
        if role == Qt.ItemDataRole.AccessibleTextRole:
            return stroke.rtfcre
        return None

    def reset(self):
        self.modelAboutToBeReset.emit()
        self._all_keys = "".join(key.strip("-") for key in system.KEYS)
        self._all_keys_filler = [" " * wcwidth(k) for k in self._all_keys]
        self._numbers = set(system.NUMBERS.values())
        self._stroke_list.clear()
        self.modelReset.emit()
        return self._all_keys

    def append(self, stroke):
        row = len(self._stroke_list)
        self.beginInsertRows(QModelIndex(), row, row)
        self._stroke_list.append(stroke)
        self.endInsertRows()

    def mimeTypes(self):
        return ["text/plain"]

    def mimeData(self, indexes):
        data = QMimeData()
        data.setText(
            "\n".join(
                filter(
                    None,
                    (
                        self.data(index, Qt.ItemDataRole.DisplayRole)
                        for index in indexes
                    ),
                )
            )
        )
        return data


class PaperTape(Tool, Ui_PaperTape):
    # i18n: Widget: “PaperTape”, tooltip.
    __doc__ = _("Paper tape display of strokes.")

    TITLE = _("Paper Tape")
    ICON = ":/resources/tape.svg"
    ROLE = "paper_tape"
    SHORTCUT = "Ctrl+T"

    def __init__(self, engine):
        super().__init__(engine)
        self.setupUi(self)
        self._model = TapeModel()
        self.header.setContentsMargins(4, 0, 0, 0)
        self.styles.addItems(TAPE_STYLES)
        self.tape.setModel(self._model)

        # Set a OS‑native fixed‑width (monospace) font.
        fixed_font = QFontDatabase.systemFont(QFontDatabase.SystemFont.FixedFont)
        fixed_font.setPointSize(fixed_font.pointSize() + 4)
        self.header.setFont(fixed_font)
        self.tape.setFont(fixed_font)

        self.tape.setSelectionMode(self.tape.SelectionMode.ExtendedSelection)
        self._copy_action = ActionCopyViewSelectionToClipboard(self.tape)
        self.tape.addAction(self._copy_action)
        # Toolbar.
        actions = [
            self.action_SelectFont,
            self.action_Clear,
            self.action_Save,
        ]
        if not utils.is_wayland():
            # Wayland does not support window on top.
            actions.insert(0, self.action_ToggleOnTop)

        self.layout().addWidget(ToolBar(*actions))
        self.action_Clear.setEnabled(False)
        self.action_Save.setEnabled(False)
        engine.signal_connect("config_changed", self.on_config_changed)
        self.on_config_changed(engine.config)
        engine.signal_connect("stroked", self.handle_stroke)
        self.tape.setFocus()
        self.restore_state()
        self.finished.connect(self.save_state)

    def _restore_state(self, settings):
        style = settings.value("style", None, int)
        if style is not None:
            style = TAPE_STYLES[style]
            self.styles.setCurrentText(style)
            self.change_style(style)
        font_string = settings.value("font")
        if font_string is not None:
            font = QFont()
            if font.fromString(font_string):
                self.header.setFont(font)
                self.tape.setFont(font)
        ontop = settings.value("ontop", None, bool)
        if ontop is not None:
            self.action_ToggleOnTop.setChecked(ontop)
            self.toggle_ontop(ontop)

    def _save_state(self, settings):
        settings.setValue("style", TAPE_STYLES.index(self._style))
        settings.setValue("font", self.tape.font().toString())
        ontop = bool(self.windowFlags() & Qt.WindowType.WindowStaysOnTopHint)
        settings.setValue("ontop", ontop)

    def on_config_changed(self, config):
        if "system_name" in config:
            all_keys = self._model.reset()
            self.header.setText(all_keys)

    @property
    def _scroll_at_end(self):
        scrollbar = self.tape.verticalScrollBar()
        return scrollbar.value() == scrollbar.maximum()

    @property
    def _style(self):
        return self.styles.currentText()

    @Slot(Stroke)
    def handle_stroke(self, stroke):
        scroll_at_end = self._scroll_at_end
        self._model.append(stroke)
        if scroll_at_end:
            self.tape.scrollToBottom()
        self.action_Clear.setEnabled(True)
        self.action_Save.setEnabled(True)

    @Slot(str)
    def change_style(self, style):
        assert style in TAPE_STYLES
        scroll_at_end = self._scroll_at_end
        self._model.style = style
        self.header.setVisible(style == STYLE_PAPER)
        if scroll_at_end:
            self.tape.scrollToBottom()

    @Slot()
    def select_font(self):
        ok, font = QFontDialog.getFont(self.tape.font(), self, "")
        if ok:
            self.header.setFont(font)
            self.tape.setFont(font)

    @Slot(bool)
    def toggle_ontop(self, ontop):
        self.setWindowFlag(Qt.WindowStaysOnTopHint, ontop)
        self.show()

    @Slot()
    def clear(self):
        flags = self.windowFlags()
        msgbox = QMessageBox()
        msgbox.setText(_("Do you want to clear the paper tape?"))
        msgbox.setStandardButtons(QMessageBox.Yes | QMessageBox.No)
        # Make sure the message box ends up above the paper tape!
        msgbox.setWindowFlags(
            msgbox.windowFlags() | (flags & Qt.WindowType.WindowStaysOnTopHint)
        )
        if QMessageBox.Yes != msgbox.exec():
            return
        self._strokes = []
        self.action_Clear.setEnabled(False)
        self.action_Save.setEnabled(False)
        self._model.reset()

    @Slot()
    def save(self):
        filename_suggestion = "steno-notes-%s.txt" % time.strftime("%Y-%m-%d-%H-%M")
        filename = QFileDialog.getSaveFileName(
            self,
            _("Save Paper Tape"),
            filename_suggestion,
            # i18n: Paper tape, "save" file picker.
            _("Text files (*.txt)"),
        )[0]
        if not filename:
            return
        with open(filename, "w") as fp:
            for row in range(self._model.rowCount(self._model.index(-1, -1))):
                print(
                    self._model.data(
                        self._model.index(row, 0), Qt.ItemDataRole.DisplayRole
                    ),
                    file=fp,
                )



#!/usr/bin/env python3

import os
import subprocess
import sys


# Default directory for caching wheels.
WHEELS_CACHE = os.path.join(".cache", "wheels")


def _split_opts(text):
    args = text.split()
    assert (len(args) % 2) == 0
    return {name: int(nb_args) for name, nb_args in zip(*([iter(args)] * 2))}


# Allowed `pip install/wheel` options.
_PIP_OPTS = _split_opts(
    # General.
    """
    --isolated 0
    -v 0 --verbose 0
    -q 0 --quiet 0
    --log 1
    --proxy 1
    --retries 1
    --timeout 1
    --exists-action 1
    --trusted-host 1
    --cert 1
    --client-cert 1
    --cache-dir 1
    --no-cache-dir 0
    --disable-pip-version-check 0
    --progress-bar 1
    """
    # Install/Wheel.
    """
    -c 1 --constraint 1
    -r 1 --requirement 1
    --no-deps 0
    --use-pep517 0
    """
    # Package Index.
    """
    -i 1 --index-url 1
    --extra-index-url 1
    --no-index 0
    -f 1 --find-links 1
    --process-dependency-links 0
    """
)

# Allowed `pip install` only options.
_PIP_INSTALL_OPTS = _split_opts(
    """
    -t 1 --target 1
    -U 0 --upgrade 0
    --upgrade-strategy 1
    --force-reinstall 0
    -I 0 --ignore-installed 0
    --user 0
    --root 1
    --prefix 1
    --require-hashes 0
    --no-warn-script-location 0
    """
)


def _pip(args, pip_install=None, verbose=True, no_progress=True):
    if no_progress:
        args.append("--progress-bar=off")
    cmd = [sys.executable]
    if sys.flags.no_site:
        cmd.append("-S")
    if sys.flags.no_user_site:
        cmd.append("-s")
    if sys.flags.ignore_environment:
        cmd.append("-E")
    if pip_install is not None:
        cmd.append(pip_install)
    else:
        cmd.extend(("-m", "pip"))
    cmd.extend(args)
    if verbose:
        print("running", " ".join(cmd), flush=True)
    return subprocess.call(cmd)


def install_wheels(
    args, pip_install=None, verbose=True, no_install=False, no_progress=True
):
    wheels_cache = WHEELS_CACHE
    wheel_args = []
    install_args = []
    while len(args) > 0:
        a = args.pop(0)
        if not a.startswith("-"):
            wheel_args.append(a)
            install_args.append(a)
            continue
        if "=" in a:
            a = a.split("=", 1)
            args.insert(0, a[1])
            a = a[0]
        opt = a
        if opt == "-w":
            wheels_cache = args.pop(0)
            continue
        if opt == "--no-install":
            no_install = True
            continue
        if opt == "--progress-bar":
            if args[0] == "off":
                no_progress = True
                del args[0]
                continue
            no_progress = False
        if opt in _PIP_OPTS:
            nb_args = _PIP_OPTS[opt]
            install_only = False
        elif opt in _PIP_INSTALL_OPTS:
            nb_args = _PIP_INSTALL_OPTS[opt]
            install_only = True
        else:
            raise ValueError("unsupported option: %s" % opt)
        a = [opt] + args[:nb_args]
        del args[:nb_args]
        if not install_only:
            wheel_args.extend(a)
        install_args.extend(a)
    wheel_args[0:0] = ["wheel", "-f", wheels_cache, "-w", wheels_cache]
    install_args[0:0] = ["install", "--no-index", "--no-cache-dir", "-f", wheels_cache]
    pip_kwargs = dict(pip_install=pip_install, no_progress=no_progress)
    code = _pip(wheel_args, **pip_kwargs)
    if code == 0 and not no_install:
        code = _pip(install_args, **pip_kwargs)
    if code != 0:
        raise Exception("wheels installation failed: pip execution returned %u" % code)


if __name__ == "__main__":
    install_wheels(sys.argv[1:])



# Copyright (c) 2010-2011 Joshua Harlan Lifton.
# See LICENSE.txt for details.

"""Plover: Open Source Stenography Software"""

if __name__ == "plover":
    from plover.i18n import Translator

    _ = Translator(__package__)
else:
    # exec from `setup.py`, package data
    # may not be available, and we don't
    # want to translate anyway.
    _ = lambda s: s

__version__ = "5.1.0"
__copyright__ = "(C) Open Steno Project"
__url__ = "http://www.openstenoproject.org/"
__download_url__ = "http://www.openstenoproject.org/plover"
__credits__ = _("""\
Founded by stenographer Mirabai Knight.

Developers:

Joshua Lifton
Hesky Fisher
Ted Morin
Benoit Pierre
Sammi Ta
Martin Koerner

and many more on GitHub:
<https://github.com/openstenoproject/plover>""")
__license__ = "GNU General Public License v2 or later (GPLv2+)"
# i18n: Short description for Plover, currently not used in the interface.
__description__ = _("Open Source Stenography Software")
__long_description__ = _("""\
Plover is a free open source program intended to bring realtime
stenographic technology not just to stenographers, but also to
hackers, hobbyists, accessibility mavens, and all-around speed demons.""")



#!/usr/bin/env python3

from pathlib import Path
import functools
import operator
import os.path
import stat
import sys


BLOCK_SIZES = (
    (1024 * 1024 * 1024 * 1024, "T"),
    (1024 * 1024 * 1024, "G"),
    (1024 * 1024, "M"),
    (1024, "K"),
)


def format_size(size):
    for bs, unit in BLOCK_SIZES:
        if size >= bs:
            return "%.1f%s" % (size / bs, unit)
    return str(size)


def tree(path, dirs_only=False, max_depth=0, _depth=0):
    path = Path(path)
    lst = path.lstat()
    is_symlink = stat.S_ISLNK(lst.st_mode)
    st = lst if is_symlink else path.stat()
    is_dir = stat.S_ISDIR(st.st_mode)
    if is_symlink:
        size = 0
    elif is_dir:
        size = functools.reduce(
            operator.add,
            [
                tree(p, dirs_only=dirs_only, max_depth=max_depth, _depth=_depth + 1)
                for p in sorted(path.iterdir())
            ],
            0,
        )
    else:
        size = lst.st_size
    if (is_dir or not dirs_only) and (not max_depth or _depth <= max_depth):
        p = str(path)
        if is_dir:
            p += os.path.sep
        if is_symlink:
            p += " -> " + os.readlink(str(path))
        print("%10s  %s" % (format_size(size), p))
    return size


def main():
    args = []
    max_depth = 0
    dirs_only = False
    argv = iter(sys.argv[1:])
    for a in argv:
        if a == "-d":
            dirs_only = True
            continue
        if a == "-L":
            max_depth = int(next(argv))
            continue
        if a.startswith("-"):
            raise ValueError(a)
        args.append(a)
    for a in args:
        tree(a, dirs_only=dirs_only, max_depth=max_depth)


if __name__ == "__main__":
    main()



from PySide6.QtGui import QImage, QTextDocument
import re
from PySide6.QtWidgets import QTextBrowser
from PySide6.QtCore import QUrl, Signal

from plover.plugins_manager.requests import CachedSession, CachedFuturesSession

from plover import log


class InfoBrowser(QTextBrowser):
    _resource_downloaded = Signal(str, bytes)

    def __init__(self, parent=None):
        super().__init__(parent=parent)
        # Regex to remove any explicit font-family attribute
        self._font_family_re = re.compile(
            r'\s*font-family\s*=\s*["\'][^"\']*["\']', re.IGNORECASE
        )
        self.setOpenExternalLinks(True)
        self._images = {}
        self._session = CachedSession()
        self._futures_session = None
        self._resource_downloaded.connect(self._update_image_resource)

    def loadResource(self, resource_type, resource_url):
        if resource_url.isLocalFile():
            resource = super().loadResource(resource_type, resource_url)
        else:
            resource = None
        resource_url = resource_url.url()
        if (
            resource is None
            and resource_type == QTextDocument.ResourceType.ImageResource
            and resource_url not in self._images
        ):
            future = self._futures_session.get(resource_url)
            future.add_done_callback(self._request_finished)
            self._images[resource_url] = future
        return resource

    def _request_finished(self, future):
        if not future.done():
            return
        try:
            resp = future.result()
        except Exception as exc:
            log.error("error fetching %s", exc.request.url, exc_info=True)
            return
        if not resp.ok:
            log.info("error fetching %s: %s", resp.url, resp.reason)
            return
        self._resource_downloaded.emit(resp.request.url, resp.content)

    def _reset_session(self):
        self._images.clear()
        if self._futures_session is not None:
            self._futures_session.close()
        self._futures_session = CachedFuturesSession(session=self._session)

    def setHtml(self, html):
        self._reset_session()
        super().setHtml(html)

    def setSource(self, url, kind):
        self._reset_session()
        super().setSource(url, kind)

    def _iter_fragments(self):
        bl = self.document().firstBlock()
        while bl.blockNumber() != -1:
            i = bl.begin()
            while not i.atEnd():
                frag = i.fragment()
                if frag.isValid():
                    yield frag
                i += 1
            bl = bl.next()

    def _sanitize_svg(self, data: bytes, url: str = "") -> bytes:
        """
        If *data* represents an SVG image, remove any explicit ``font-family``
        attributes so Qt will not attempt to resolve fonts that might be
        missing on the host system (which triggers slow alias‑population
        scans and warnings).

        Detection is done by:
        1.  URL ending with ``.svg`` **or**
        2.  Looking for ``<svg`` within the first ~256 decoded bytes.

        The function returns the (possibly modified) byte string.
        """
        # try to detect SVG by extension or content sniff.
        is_svg = url.lower().endswith(".svg")
        if not is_svg:
            try:
                head = data[:256].decode("utf-8", "ignore").lstrip().lower()
                is_svg = head.startswith("<svg") or "<svg" in head[:20]
            except Exception:
                # If decoding fails, assume it's not SVG.
                return data

        if not is_svg:
            return data

        # Strip any font-family attribute.
        try:
            text = data.decode("utf-8", "replace")
            text = self._font_family_re.sub("", text)
            return text.encode("utf-8")
        except Exception:
            # On failure, return the original bytes unchanged.
            return data

    def _update_image_resource(self, url, data):
        if url not in self._images:
            # Ignore request from a previous document.
            return

        data = self._sanitize_svg(data, url)

        image = QImage.fromData(data)
        if image is None:
            log.warning("could not load image from %s", url)
            return
        doc = self.document()
        doc.addResource(QTextDocument.ResourceType.ImageResource, QUrl(url), image)
        for frag in self._iter_fragments():
            fmt = frag.charFormat()
            if fmt.isImageFormat() and fmt.toImageFormat().name() == url:
                doc.markContentsDirty(frag.position(), frag.length())



# Copyright (c) 2016 Open Steno Project
# See LICENSE.txt for details.

import os

from plover.oslayer.config import CONFIG_DIR
from plover.resource import ASSET_SCHEME


def popcount_8(v):
    """Population count for an 8 bit integer"""
    assert 0 <= v <= 255
    v -= (v >> 1) & 0x55555555
    v = (v & 0x33333333) + ((v >> 2) & 0x33333333)
    return ((v + (v >> 4) & 0xF0F0F0F) * 0x1010101) >> 24


def expand_path(path):
    """Expand path.

    - if starting with "~/", it is substituted with the user home directory
    - if relative, it is resolved relative to CONFIG_DIR
    """
    if path.startswith(ASSET_SCHEME):
        return path
    path = os.path.expanduser(path)
    path = normalize_path(os.path.join(CONFIG_DIR, path))
    return path


def shorten_path(path):
    """Shorten path.

    - if the path is below CONFIG_DIR, a relative path to it is returned
    - if path is below the user home directory, "~/" is substituted to it

    Note: relative path are automatically assumed to be relative to CONFIG_DIR.
    """
    if path.startswith(ASSET_SCHEME):
        return path
    path = normalize_path(os.path.join(CONFIG_DIR, path))
    config_dir = normalize_path(CONFIG_DIR)
    if not config_dir.endswith(os.sep):
        config_dir += os.sep
    if path.startswith(config_dir):
        return path[len(config_dir) :]
    home_dir = normalize_path(os.path.expanduser("~"))
    if not home_dir.endswith(os.sep):
        home_dir += os.sep
    if path.startswith(home_dir):
        return os.path.join("~", path[len(home_dir) :])
    return path


def normalize_path(path):
    """Normalize path: return canonical path, normalizing case on Windows."""
    if path.startswith(ASSET_SCHEME):
        return path
    return os.path.normcase(os.path.realpath(path))


def boolean(value):
    if isinstance(value, str):
        v = value.lower()
        if v in ("1", "yes", "true", "on"):
            return True
        if v in ("0", "no", "false", "off"):
            return False
        raise ValueError(value)
    return bool(value)


def to_surrogate_pair(char):
    pairs = []
    for code in char:
        code_point = ord(code)
        if code_point >= 0x10000:
            high_part = (code_point - 0x10000) // 0x400 + 0xD800
            low_part = (code_point - 0x10000) % 0x400 + 0xDC00
            pairs += (high_part, low_part)
        else:
            pairs.append(code_point)
    return pairs



from plyer import notification
import logging
import os

from plover import log, __name__ as __software_name__
from plover.oslayer.config import ASSETS_DIR

APPNAME = __software_name__.capitalize()
APPICON = os.path.join(ASSETS_DIR, "plover.ico")

# Windows NOTIFYICONDATAW limits
_MAX_TITLE = 60  # leave some margin (spec is 64)
_MAX_BODY = 250  # leave some margin (spec is 256)


def _flatten(s: str) -> str:
    # Remove newlines/tabs that can expand length; collapse whitespace.
    return " ".join(s.replace("\r", " ").replace("\n", " ").replace("\t", " ").split())


def _truncate(s: str, limit: int) -> str:
    s = _flatten(s)
    if len(s) <= limit:
        return s
    # keep room for ellipsis
    if limit <= 1:
        return s[:limit]
    return s[: limit - 1] + "…"


class NotificationHandler(logging.Handler):
    def __init__(self):
        super().__init__()
        self.setLevel(log.WARNING)
        self.setFormatter(
            log.NoExceptionTracebackFormatter("%(levelname)s: %(message)s")
        )

    def handle(self, record):
        level = record.levelno
        message = self.format(record)

        title = _truncate(APPNAME, _MAX_TITLE)
        body = _truncate(message, _MAX_BODY)

        # Reasonable timeouts
        if level <= log.INFO:
            timeout = 10
        elif level <= log.WARNING:
            timeout = 15
        else:
            timeout = 60

        notification.notify(
            app_name=title,
            app_icon=APPICON,
            title=title,
            message=body,
            timeout=timeout,
        )



import re
import sys
from importlib.metadata import entry_points

if __name__ == "__main__":
    if sys.platform.startswith("win32"):
        # Allow interrupting with Ctrl+C.
        __import__("ctypes").windll.kernel32.SetConsoleCtrlHandler(0, 0)

    # Normalize the script name in `sys.argv[0]`
    sys.argv[0] = re.sub(r"(-script\.pyw?|\.exe)?$", "", sys.argv[0])

    # Find and load the `pip` entry point
    entry_point = next(
        ep for ep in entry_points(group="console_scripts") if ep.name == "pip"
    )
    sys.exit(entry_point.load()())



# -*- coding: utf-8 -*-
# Copyright (c) 2010 Joshua Harlan Lifton.
# See LICENSE.txt for details.

"For use with a computer keyboard (preferably NKRO) as a steno machine."

from plover import _
from plover.machine.base import StenotypeBase
from plover.misc import boolean
from plover.oslayer.keyboardcontrol import KeyboardCapture


# i18n: Machine name.
_._("Keyboard")


class Keyboard(StenotypeBase):
    """Standard stenotype interface for a computer keyboard.

    This class implements the three methods necessary for a standard
    stenotype interface: start_capture, stop_capture, and
    add_callback.

    """

    KEYS_LAYOUT = """
    Escape  F1 F2 F3 F4  F5 F6 F7 F8  F9 F10 F11 F12

      `  1  2  3  4  5  6  7  8  9  0  -  =  \\ BackSpace  Insert Home Page_Up
     Tab  q  w  e  r  t  y  u  i  o  p  [  ]               Delete End  Page_Down
           a  s  d  f  g  h  j  k  l  ;  '      Return
            z  x  c  v  b  n  m  ,  .  /                          Up
                     space                                   Left Down Right
    """
    ACTIONS = ("arpeggiate",)

    def __init__(self, params):
        """Monitor the keyboard's events."""
        super().__init__()
        self._arpeggiate = params["arpeggiate"]
        self._first_up_chord_send = params["first_up_chord_send"]
        if self._arpeggiate and self._first_up_chord_send:
            self._error()
            raise RuntimeError(
                "Arpeggiate and first-up chord send cannot both be enabled!"
            )
        self._is_suppressed = False
        # Currently held keys.
        self._down_keys = set()
        if self._first_up_chord_send:
            # If this is True, the first key in a stroke has already been released
            # and subsequent key-up events should not send more strokes
            self._chord_already_sent = False
        else:
            # Collect the keys in the stroke, in case first_up_chord_send is False
            self._stroke_keys = set()
        self._keyboard_capture = None
        self._last_stroke_key_down_count = 0
        self._stroke_key_down_count = 0
        self._update_bindings()

    def _update_suppression(self):
        if self._keyboard_capture is None:
            return
        suppressed_keys = self._bindings.keys() if self._is_suppressed else ()
        self._keyboard_capture.suppress(suppressed_keys)

    def _update_bindings(self):
        self._arpeggiate_key = None
        self._bindings = dict(self.keymap.get_bindings())
        for key, mapping in list(self._bindings.items()):
            if "no-op" == mapping:
                self._bindings[key] = None
            elif "arpeggiate" == mapping:
                if self._arpeggiate:
                    self._bindings[key] = None
                    self._arpeggiate_key = key
                else:
                    # Don't suppress arpeggiate key if it's not used.
                    del self._bindings[key]
        self._update_suppression()

    def set_keymap(self, keymap):
        super().set_keymap(keymap)
        self._update_bindings()

    def start_capture(self):
        """Begin listening for output from the stenotype machine."""
        self._initializing()
        try:
            self._keyboard_capture = KeyboardCapture()
            self._keyboard_capture.key_down = self._key_down
            self._keyboard_capture.key_up = self._key_up
            self._keyboard_capture.start()
            self._update_suppression()
        except:
            self._error()
            raise
        self._ready()

    def stop_capture(self):
        """Stop listening for output from the stenotype machine."""
        if self._keyboard_capture is not None:
            self._is_suppressed = False
            self._update_suppression()
            self._keyboard_capture.cancel()
            self._keyboard_capture = None
        self._stopped()

    def set_suppression(self, enabled):
        self._is_suppressed = enabled
        self._update_suppression()

    def suppress_last_stroke(self, send_backspaces):
        send_backspaces(self._last_stroke_key_down_count)
        self._last_stroke_key_down_count = 0

    def _key_down(self, key):
        """Called when a key is pressed."""
        assert key is not None
        self._stroke_key_down_count += 1
        self._down_keys.add(key)
        if self._first_up_chord_send:
            self._chord_already_sent = False
        else:
            self._stroke_keys.add(key)

    def _key_up(self, key):
        """Called when a key is released."""
        assert key is not None

        self._down_keys.discard(key)

        if self._first_up_chord_send:
            if self._chord_already_sent:
                return
        else:
            # A stroke is complete if all pressed keys have been released,
            # and — when arpeggiate mode is enabled — the arpeggiate key
            # is part of it.
            if (
                self._down_keys
                or not self._stroke_keys
                or (self._arpeggiate and self._arpeggiate_key not in self._stroke_keys)
            ):
                return

        self._last_stroke_key_down_count = self._stroke_key_down_count
        if self._first_up_chord_send:
            steno_keys = {self._bindings.get(k) for k in self._down_keys | {key}}
            self._chord_already_sent = True
        else:
            steno_keys = {self._bindings.get(k) for k in self._stroke_keys}
            self._stroke_keys.clear()
        steno_keys -= {None}
        if steno_keys:
            self._notify(steno_keys)
        self._stroke_key_down_count = 0

    @classmethod
    def get_option_info(cls):
        return {
            "arpeggiate": (False, boolean),
            "first_up_chord_send": (False, boolean),
        }



# Copyright (c) 2010-2011 Joshua Harlan Lifton.
# See LICENSE.txt for details.

"""Functions that implement some English orthographic rules."""

from plover import system


def make_candidates_from_rules(word, suffix, check=lambda x: True):
    candidates = []
    for r in system.ORTHOGRAPHY_RULES:
        m = r[0].match(word + " ^ " + suffix)
        if m:
            expanded = m.expand(r[1])
            if check(expanded):
                candidates.append(expanded)
    return candidates


def _add_suffix(word, suffix):
    in_dict_f = lambda x: x in system.ORTHOGRAPHY_WORDS

    candidates = []

    alias = system.ORTHOGRAPHY_RULES_ALIASES.get(suffix, None)
    if alias is not None:
        candidates.extend(make_candidates_from_rules(word, alias, in_dict_f))

    # Try a simple join if it is in the dictionary.
    simple = word + suffix
    if in_dict_f(simple):
        candidates.append(simple)

    # Try rules with dict lookup.
    candidates.extend(make_candidates_from_rules(word, suffix, in_dict_f))

    # For all candidates sort by prominence in dictionary and, since sort is
    # stable, also by the order added to candidates list.
    if candidates:
        candidates.sort(key=lambda x: system.ORTHOGRAPHY_WORDS[x])
        return candidates[0]

    # Try rules without dict lookup.
    candidates = make_candidates_from_rules(word, suffix)
    if candidates:
        return candidates[0]

    # If all else fails then just do a simple join.
    return simple


def add_suffix(word, suffix):
    """Add a suffix to a word by applying the rules above

    Arguments:

    word -- A word
    suffix -- The suffix to add

    """
    suffix, sep, rest = suffix.partition(" ")
    expanded = _add_suffix(word, suffix)
    return expanded + sep + rest



# -*- coding: utf-8 -*-
# Copyright (c) 2013 Hesky Fisher
# See LICENSE.txt for details.

"""Unit tests for json.py."""

from plover.dictionary.json_dict import JsonDictionary

from plover_build_utils.testing import dictionary_test


def json_load_test(contents, expected):
    if isinstance(contents, str):
        contents = contents.encode("utf-8")
    return contents, expected


JSON_LOAD_TESTS = (
    lambda: json_load_test('{"S": "a"}', '"S": "a"'),
    # Default encoding is utf-8.
    lambda: json_load_test('{"S": "café"}', '"S": "café"'),
    # But if that fails, the implementation
    # must automatically retry with latin-1.
    lambda: json_load_test('{"S": "café"}'.encode("latin-1"), '"S": "café"'),
    # Invalid JSON.
    lambda: json_load_test('{"foo", "bar",}', ValueError),
    # Invalid JSON.
    lambda: json_load_test("foo", ValueError),
    # Cannot convert to dict.
    lambda: json_load_test('"foo"', ValueError),
    # Ditto.
    lambda: json_load_test("4.2", TypeError),
)


def json_save_test(entries, expected):
    return entries, expected.encode("utf-8")


JSON_SAVE_TESTS = (
    # Simple test.
    lambda: json_save_test(
        """
        'S': 'a',
        """,
        '{\n"S": "a"\n}\n',
    ),
    # Check strokes format: '/' separated.
    lambda: json_save_test(
        """
        'SAPL/-PL': 'sample',
        """,
        '{\n"SAPL/-PL": "sample"\n}\n',
    ),
    # Contents should be saved as UTF-8, no escaping.
    lambda: json_save_test(
        """
        'S': 'café',
        """,
        '{\n"S": "café"\n}\n',
    ),
    # Check escaping of special characters.
    lambda: json_save_test(
        r"""
        'S': '{^"\n\t"^}',
        """,
        '{\n"S": "' + r"{^\"\n\t\"^}" + '"\n}\n',
    ),
    # Keys are sorted on save.
    lambda: json_save_test(
        """
        'T': 'bravo',
        'S': 'alpha',
        'R': 'charlie',
        """,
        '{\n"S": "alpha",\n"T": "bravo",\n"R": "charlie"\n}\n',
    ),
)


@dictionary_test
class TestJsonDictionary:
    DICT_CLASS = JsonDictionary
    DICT_EXTENSION = "json"
    DICT_REGISTERED = True
    DICT_LOAD_TESTS = JSON_LOAD_TESTS
    DICT_SAVE_TESTS = JSON_SAVE_TESTS
    DICT_SAMPLE = b"{}"



# Copyright (c) 2013 Hesky Fisher
# See LICENSE.txt for details.

"""Tests for loading_manager.py."""

from collections import defaultdict
import os
import tempfile

import pytest

from plover.engine import ErroredDictionary
import plover.dictionary.loading_manager as loading_manager


class FakeDictionaryContents:
    def __init__(self, contents, timestamp):
        self.contents = contents
        self.timestamp = timestamp

    def __eq__(self, other):
        if isinstance(other, FakeDictionaryContents):
            return self.contents == other.contents
        return self.contents == other


class FakeDictionaryInfo:
    def __init__(self, name, contents):
        self.name = name
        self.contents = contents
        self.tf = tempfile.NamedTemporaryFile()

    def __repr__(self):
        return "FakeDictionaryInfo(%r, %r)" % (self.name, self.contents)


class MockLoader:
    def __init__(self, files):
        self.files = files
        self.load_counts = defaultdict(int)

    def __call__(self, filename):
        self.load_counts[filename] += 1
        d = self.files[filename]
        if isinstance(d.contents, Exception):
            raise d.contents
        timestamp = os.path.getmtime(filename)
        return FakeDictionaryContents(d.contents, timestamp)


def test_loading(monkeypatch):
    dictionaries = {}
    for i in range(8):
        c = chr(ord("a") + i)
        contents = c * 5
        if i >= 4:
            contents = Exception(contents)
        d = FakeDictionaryInfo(c, contents)
        assert d.tf.name not in dictionaries
        dictionaries[d.tf.name] = d
        assert c not in dictionaries
        dictionaries[c] = d

    def df(name):
        return dictionaries[name].tf.name

    loader = MockLoader(dictionaries)
    monkeypatch.setattr("plover.dictionary.loading_manager.load_dictionary", loader)
    manager = loading_manager.DictionaryLoadingManager(lambda filename, result: None)
    manager.start_loading(df("a")).get()
    manager.start_loading(df("b")).get()
    results = manager.load([df("c"), df("b")])
    # Returns the right values in the right order.
    assert results == ["ccccc", "bbbbb"]
    # Dropped superfluous files.
    assert sorted([df("b"), df("c")]) == sorted(manager.dictionaries.keys())
    # Check dict like interface.
    assert len(manager) == 2
    assert df("a") not in manager
    with pytest.raises(KeyError):
        manager[df("a")]
    assert df("b") in manager
    assert df("c") in manager
    assert results == [manager[df("c")], manager[df("b")]]
    # Return a ErroredDictionary for load errors.
    results = manager.load([df("c"), df("e"), df("b"), df("f")])
    assert len(results) == 4
    assert results[0] == "ccccc"
    assert results[2] == "bbbbb"
    assert isinstance(results[1], ErroredDictionary)
    assert results[1].path == df("e")
    assert isinstance(results[1].exception, Exception)
    assert isinstance(results[3], ErroredDictionary)
    assert results[3].path == df("f")
    assert isinstance(results[3].exception, Exception)
    # Only loaded the files once.
    assert all(x == 1 for x in loader.load_counts.values())
    # No reload if file timestamp is unchanged, or the dictionary
    # timestamp is more recent. (use case: dictionary edited with
    # Plover and saved back)
    file_timestamp = results[0].timestamp
    results[0].timestamp = file_timestamp + 1
    dictionaries["c"].contents = "CCCCC"
    results = manager.load([df("c")])
    assert results == ["ccccc"]
    assert loader.load_counts[df("c")] == 1
    # Check outdated dictionaries are reloaded.
    results[0].timestamp = file_timestamp - 1
    results = manager.load([df("c")])
    assert results == ["CCCCC"]
    assert loader.load_counts[df("c")] == 2
    # Check trimming of outdated dictionaries.
    results[0].timestamp = file_timestamp - 1
    manager.unload_outdated()
    assert len(manager) == 0
    assert df("c") not in manager



# Copyright (c) 2013 Hesky Fisher.
# See LICENSE.txt for details.

"""StenoDictionary class and related functions.

A steno dictionary maps sequences of steno strokes to translations.

"""

import collections
import os

from plover.resource import (
    ASSET_SCHEME,
    resource_filename,
    resource_timestamp,
    resource_update,
)


class StenoDictionary:
    """A steno dictionary.

    This dictionary maps immutable sequences to translations and tracks the
    length of the longest key.

    Attributes:
    longest_key -- A read only property holding the length of the longest key.
    timestamp -- File last modification time, used to detect external changes.

    """

    # False if class support creation.
    readonly = False

    def __init__(self):
        self._dict = {}
        self._longest_key = 0
        self.reverse = collections.defaultdict(list)
        # Case-insensitive reverse dict
        self.casereverse = collections.defaultdict(list)
        self.filters = []
        self.timestamp = 0
        self.readonly = False
        self.enabled = True
        self.path = None

    def __str__(self):
        return "%s(%r)" % (self.__class__.__name__, self.path)

    def __repr__(self):
        return str(self)

    @classmethod
    def create(cls, resource):
        assert not resource.startswith(ASSET_SCHEME)
        if cls.readonly:
            raise ValueError("%s does not support creation" % cls.__name__)
        d = cls()
        d.path = resource
        return d

    @classmethod
    def load(cls, resource):
        filename = resource_filename(resource)
        timestamp = resource_timestamp(filename)
        d = cls()
        d._load(filename)
        if (
            cls.readonly
            or resource.startswith(ASSET_SCHEME)
            or not os.access(filename, os.W_OK)
        ):
            d.readonly = True
        d.path = resource
        d.timestamp = timestamp
        return d

    def save(self):
        assert not self.readonly
        with resource_update(self.path) as temp_path:
            self._save(temp_path)
        self.timestamp = resource_timestamp(self.path)

    def _load(self, filename):
        raise NotImplementedError()

    def _save(self, filename):
        raise NotImplementedError()

    @property
    def longest_key(self):
        """The length of the longest key in the dict."""
        return self._longest_key

    def __len__(self):
        return self._dict.__len__()

    def __iter__(self):
        return self._dict.__iter__()

    def __getitem__(self, key):
        return self._dict.__getitem__(key)

    def clear(self):
        assert not self.readonly
        self._dict.clear()
        self.reverse.clear()
        self.casereverse.clear()
        self._longest_key = 0

    def items(self):
        return self._dict.items()

    def update(self, *args, **kwargs):
        assert not self.readonly
        iterable_list = [
            a.items() if isinstance(a, (dict, StenoDictionary)) else a for a in args
        ]
        if kwargs:
            iterable_list.append(kwargs.items())
        if not self._dict:
            reverse = self.reverse
            casereverse = self.casereverse
            longest_key = self._longest_key
            assert not (reverse or casereverse or longest_key)
            self._dict = dict(*iterable_list)
            for key, value in self._dict.items():
                reverse[value].append(key)
                casereverse[value.lower()].append(value)
                key_len = len(key)
                if key_len > longest_key:
                    longest_key = key_len
            self._longest_key = longest_key
        else:
            for iterable in iterable_list:
                for key, value in iterable:
                    self[key] = value

    def __setitem__(self, key, value):
        assert not self.readonly
        if key in self:
            del self[key]
        self._longest_key = max(self._longest_key, len(key))
        self._dict[key] = value
        self.reverse[value].append(key)
        self.casereverse[value.lower()].append(value)

    def get(self, key, fallback=None):
        return self._dict.get(key, fallback)

    def __delitem__(self, key):
        assert not self.readonly
        value = self._dict.pop(key)
        self.reverse[value].remove(key)
        self.casereverse[value.lower()].remove(value)
        if len(key) == self.longest_key:
            if self._dict:
                self._longest_key = max(len(x) for x in self._dict)
            else:
                self._longest_key = 0

    def __contains__(self, key):
        return self.get(key) is not None

    def reverse_lookup(self, value):
        return set(self.reverse.get(value, ()))

    def casereverse_lookup(self, value):
        return set(self.casereverse.get(value, ()))


class StenoDictionaryCollection:
    def __init__(self, dicts=[]):
        self.dicts = []
        self.filters = []
        self.set_dicts(dicts)

    @property
    def longest_key(self):
        return max((d.longest_key for d in self.dicts if d.enabled), default=0)

    def set_dicts(self, dicts):
        self.dicts = dicts[:]

    def _lookup_keep_deleted(self, key, dicts=None, filters=()):
        """
        Lookup a key in the given dicts.
        If key appears in none of the dicts, return None.
        If key appears in some dicts but with "{plover:deleted}" as value,
        return "{plover:deleted}".
        """
        if dicts is None:
            dicts = self.dicts
        key_len = len(key)
        if key_len > self.longest_key:
            return None
        for d in dicts:
            if not d.enabled:
                continue
            if key_len > d.longest_key:
                continue
            value = d.get(key)
            if value is not None:
                if not any(f(key, value) for f in filters):
                    return value

    def _lookup(self, key, dicts=None, filters=()):
        """
        Same as _lookup_keep_deleted, but if key appears in some dicts
        but with "{plover:deleted}" as value (case-insensitive),
        return None instead.
        """
        result = self._lookup_keep_deleted(key, dicts, filters)
        if result is None or result.lower() == "{plover:deleted}":
            return None
        return result

    def _lookup_from_all(self, key, dicts=None, filters=()):
        """Key lookup from all dictionaries

        Returns list of (value, dictionary) tuples.
        Entries with value "{plover:deleted}" are kept.
        """
        if dicts is None:
            dicts = self.dicts
        key_len = len(key)
        if key_len > self.longest_key:
            return None
        values = []
        for d in dicts:
            if not d.enabled:
                continue
            if key_len > d.longest_key:
                continue
            value = d.get(key)
            if value:
                if not any(f(key, value) for f in filters):
                    values.append((value, d))
        return values

    def __str__(self):
        return "StenoDictionaryCollection" + repr(tuple(self.dicts))

    def __repr__(self):
        return str(self)

    def lookup(self, key):
        return self._lookup(key, filters=self.filters)

    def raw_lookup(self, key):
        return self._lookup(key)

    def lookup_from_all(self, key):
        return self._lookup_from_all(key, filters=self.filters)

    def raw_lookup_from_all(self, key):
        return self._lookup_from_all(key)

    def reverse_lookup(self, value):
        keys = set()
        for n, d in enumerate(self.dicts):
            if not d.enabled:
                continue
            # Ignore key if it's overridden by a higher priority dictionary.
            keys.update(
                k
                for k in d.reverse_lookup(value)
                if self._lookup_keep_deleted(k, dicts=self.dicts[:n]) is None
            )
        return keys

    def casereverse_lookup(self, value):
        keys = set()
        for d in self.dicts:
            if not d.enabled:
                continue
            keys.update(d.casereverse_lookup(value))
        return keys

    def first_writable(self):
        """Return the first writable dictionary."""
        for d in self.dicts:
            if not d.readonly:
                return d
        raise KeyError("no writable dictionary")

    def set(self, key, value, path=None):
        if path is None:
            d = self.first_writable()
        else:
            d = self[path]
        d[key] = value

    def save(self, path_list=None):
        """Save the dictionaries in <path_list>.

        If <path_list> is None, all writable dictionaries are saved"""
        if path_list is None:
            dict_list = [d for d in self if not d.readonly]
        else:
            dict_list = [self[path] for path in path_list]
        for d in dict_list:
            assert not d.readonly
            d.save()

    def get(self, path):
        for d in self.dicts:
            if d.path == path:
                return d

    def __getitem__(self, path):
        d = self.get(path)
        if d is None:
            raise KeyError(repr(path))
        return d

    def __iter__(self):
        for d in self.dicts:
            yield d.path

    def add_filter(self, f):
        self.filters.append(f)

    def remove_filter(self, f):
        self.filters.remove(f)



import argparse
import sys

from plover import log
from plover.oslayer.controller import Controller


def main():
    description = "Send a command to a running Plover instance."
    parser = argparse.ArgumentParser(description=description)
    parser.add_argument(
        "-l",
        "--log-level",
        choices=["debug", "info", "warning", "error"],
        default=None,
        help="set log level",
    )
    parser.add_argument(
        "command", metavar="COMMAND{:ARGS}", type=str, help="the command to send"
    )
    args = parser.parse_args(args=sys.argv[1:])
    if args.log_level is not None:
        log.set_level(args.log_level.upper())
    log.setup_platform_handler()
    with Controller() as controller:
        if controller.is_owner:
            log.error("sending command failed: no running instance found")
            sys.exit(1)
        controller.send_command(args.command)


if __name__ == "__main__":
    main()



from .blackbox import blackbox_test
from .dict import make_dict
from .output import CaptureOutput
from .parametrize import parametrize
from .steno import steno_to_stroke
from .steno_dictionary import dictionary_test



from collections import namedtuple

from importlib.metadata import entry_points, PackageNotFoundError

from plover.oslayer.config import PLUGINS_PLATFORM
from plover import log


class Plugin:
    def __init__(self, plugin_type, name, obj):
        self.plugin_type = plugin_type
        self.name = name
        self.obj = obj
        self.__doc__ = obj.__doc__ or ""

    def __str__(self):
        return f"{self.plugin_type}:{self.name}"


PluginDistribution = namedtuple("PluginDistribution", "dist plugins")


class Registry:
    PLUGIN_TYPES = (
        "command",
        "dictionary",
        "extension",
        "gui",
        "gui.qt.machine_option",
        "gui.qt.tool",
        "machine",
        "macro",
        "meta",
        "system",
    )

    def __init__(self, suppress_errors=True):
        self._plugins = {}
        self._distributions = {}
        self._suppress_errors = suppress_errors
        for plugin_type in self.PLUGIN_TYPES:
            self._plugins[plugin_type] = {}

    def register_plugin(self, plugin_type, name, obj):
        plugin = Plugin(plugin_type, name, obj)
        self._plugins[plugin_type][name.lower()] = plugin
        return plugin

    def register_plugin_from_entrypoint(self, plugin_type, entrypoint):
        log.info("%s: %s (from %s)", plugin_type, entrypoint.name, entrypoint.group)
        try:
            obj = entrypoint.load()
        except:
            log.error(
                "error loading %s plugin: %s (from %s)",
                plugin_type,
                entrypoint.name,
                entrypoint.value,
                exc_info=True,
            )
            if not self._suppress_errors:
                raise
        else:
            plugin = self.register_plugin(plugin_type, entrypoint.name, obj)
            # Keep track of distributions providing plugins.
            dist_id = entrypoint.group
            dist = self._distributions.get(dist_id)
            if dist is None:
                dist = PluginDistribution(entrypoint.group, set())
                self._distributions[dist_id] = dist
            dist.plugins.add(plugin)

    def get_plugin(self, plugin_type, plugin_name):
        return self._plugins[plugin_type][plugin_name.lower()]

    def list_plugins(self, plugin_type):
        return sorted(self._plugins[plugin_type].values(), key=lambda p: p.name)

    def list_distributions(self):
        return [dist for _, dist in sorted(self._distributions.items())]

    def update(self):
        # Is support for the QT GUI available?
        try:
            qt_entry_points = entry_points(group="plover.gui")
            has_gui_qt = any(ep.name == "qt" for ep in qt_entry_points)
        except PackageNotFoundError:
            has_gui_qt = False
        # Register available plugins.
        for plugin_type in self.PLUGIN_TYPES:
            if plugin_type.startswith("gui.qt.") and not has_gui_qt:
                continue
            entrypoint_type = f"plover.{plugin_type}"
            for entrypoint in entry_points(group=entrypoint_type):
                if "gui_qt" in entrypoint.extras and not has_gui_qt:
                    continue
                self.register_plugin_from_entrypoint(plugin_type, entrypoint)
            if PLUGINS_PLATFORM is None:
                continue
            entrypoint_type = f"plover.{PLUGINS_PLATFORM}.{plugin_type}"
            for entrypoint in entry_points(group=entrypoint_type):
                self.register_plugin_from_entrypoint(plugin_type, entrypoint)


registry = Registry()



# Copyright (c) 2011 Hesky Fisher
# See LICENSE.txt for details.

"""Unit tests for stentura.py."""

import struct
import threading

import pytest

from plover.machine import stentura


def make_response(seq, action, error=0, p1=0, p2=0, data=None, length=None):
    if not length:
        length = 14
        if data:
            length += 2 + len(data)
    response = bytearray(14 + ((2 + len(data)) if data else 0))
    struct.pack_into("<2B5H", response, 0, 1, seq, length, action, error, p1, p2)
    struct.pack_into("<H", response, 12, stentura._crc(response, 1, 11))
    if data:
        response[14 : 14 + len(data)] = data
        struct.pack_into("<H", response, 14 + len(data), stentura._crc(data))
    return response


def make_read_response(seq, data=[]):
    return make_response(seq, stentura._READC, p1=len(data), data=data)


def make_readc_packets(data):
    requests, responses = [], []
    seq = stentura._SequenceCounter()
    buf = bytearray(256)
    block, byte = 0, 0
    while data:
        s = seq()
        chunk = data[0:512]
        data = data[512:]
        q = stentura._make_read(buf, s, block, byte)
        requests.append(bytes(q))
        r = make_read_response(s, chunk)
        responses.append(bytes(r))
        byte += len(chunk)
        if byte >= 512:
            block += 1
            byte -= 512
    s = seq()
    q = stentura._make_read(buf, s, block, byte)
    requests.append(bytes(q))
    r = make_read_response(s)
    responses.append(bytes(r))
    return requests, responses


def parse_request(request):
    header = struct.unpack_from("<2B8H", request)
    if header[2] > 18:
        header = list(header) + [request[18:-2], struct.unpack("<H", request[-2:])]
    else:
        header = list(header) + [None] * 2
    return dict(
        zip(
            [
                "SOH",
                "seq",
                "length",
                "action",
                "p1",
                "p2",
                "p3",
                "p4",
                "p5",
                "crc",
                "data",
                "data_crc",
            ],
            header,
        )
    )


class MockPacketPort:
    def __init__(self, responses, requests=None):
        self._responses = responses
        self.writes = 0
        self._requests = requests
        self._current_response_offset = None

    def write(self, data):
        self.writes += 1
        if self._requests and self._requests[self.writes - 1] != bytes(data):
            raise Exception("Wrong packet.")
        self._current_response_offset = 0
        return len(data)

    def read(self, count):
        response = self._responses[self.writes - 1]
        data = response[
            self._current_response_offset : self._current_response_offset + count
        ]
        self._current_response_offset += count
        return data


def test_crc():
    data = bytearray(b"123456789")
    assert stentura._crc(data) == 0xBB3D


def test_write_buffer():
    buf = bytearray()
    data = [1, 2, 3]
    stentura._write_to_buffer(buf, 0, data)
    assert list(buf) == data
    stentura._write_to_buffer(buf, 0, [5, 6])
    assert list(buf) == [5, 6, 3]


def test_parse_stroke():
    # SAT
    a = 0b11001000
    b = 0b11000100
    c = 0b11000000
    d = 0b11001000
    assert sorted(stentura._parse_stroke(a, b, c, d)) == sorted(["S-", "A-", "-T"])


# 11^#STKP 11WHRAO* 11EUFRPB 11LGTSDZ
# PRAOERBGS
def test_parse_strokes():
    data = bytearray()
    # SAT
    a = 0b11001000
    b = 0b11000100
    c = 0b11000000
    d = 0b11001000
    data.extend([a, b, c, d])
    # PRAOERBGS
    a = 0b11000001
    b = 0b11001110
    c = 0b11100101
    d = 0b11010100
    data.extend([a, b, c, d])
    for result, expected in zip(
        stentura._parse_strokes(bytes(data)),
        [["S-", "A-", "-T"], ["P-", "R-", "A-", "O-", "-E", "-R", "-B", "-G", "-S"]],
    ):
        assert sorted(result) == sorted(expected)


def test_make_request():
    buf = bytearray(range(256))
    seq = 2
    action = stentura._OPEN
    p1, p2, p3, p4, p5 = 1, 2, 3, 4, 5
    p = stentura._make_request(buf, action, seq, p1, p2, p3, p4, p5)
    for_crc = [seq, 18, 0, action, 0, p1, 0, p2, 0, p3, 0, p4, 0, p5, 0]
    crc = stentura._crc(for_crc)
    expected = bytearray([1] + for_crc + [crc & 0xFF, crc >> 8])
    assert p == expected
    # Now with data.
    data = b"Testing Testing 123"
    p = stentura._make_request(buf, action, seq, p1, p2, p3, p4, p5, data)
    length = 18 + len(data) + 2
    for_crc = [
        seq,
        length & 0xFF,
        length >> 8,
        action,
        0,
        p1,
        0,
        p2,
        0,
        p3,
        0,
        p4,
        0,
        p5,
        0,
    ]
    crc = stentura._crc(for_crc)
    data_crc = stentura._crc(data)
    expected = bytearray([1] + for_crc + [crc & 0xFF, crc >> 8])
    expected.extend(data)
    expected.extend([data_crc & 0xFF, data_crc >> 8])
    assert p == expected


def test_make_open():
    buf = bytearray(range(32))  # Start with junk in the buffer.
    seq = 79
    drive = b"A"
    filename = b"REALTIME.000"
    p = stentura._make_open(buf, seq, drive, filename)
    for_crc = [
        seq,
        20 + len(filename),
        0,
        stentura._OPEN,
        0,
        ord(drive),
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
    ]
    crc = stentura._crc(for_crc)
    data_crc = stentura._crc(filename)
    expected = bytearray([1] + for_crc + [crc & 0xFF, crc >> 8])
    expected.extend(filename)
    expected.extend([data_crc & 0xFF, data_crc >> 8])
    assert p == expected


def test_make_read():
    buf = bytearray(range(32))  # Start with junk in the buffer.
    seq = 32
    block = 1
    byte = 8
    length = 20
    p = stentura._make_read(buf, seq, block, byte, length)
    for_crc = [seq, 18, 0, stentura._READC, 0, 1, 0, 0, 0, length, 0, block, 0, byte, 0]
    crc = stentura._crc(for_crc)
    expected = bytearray([1] + for_crc + [crc & 0xFF, crc >> 8])
    assert p == expected


def test_make_reset():
    buf = bytearray(range(32))  # Start with junk in the buffer.
    seq = 67
    p = stentura._make_reset(buf, seq)
    for_crc = [seq, 18, 0, stentura._RESET, 0] + ([0] * 10)
    crc = stentura._crc(for_crc)
    expected = bytearray([1] + for_crc + [crc & 0xFF, crc >> 8])
    assert p == expected


VALIDATE_RESPONSE_TESTS = (
    ("valid, no data", make_response(5, 9, 1, 2, 3), True),
    ("valid, data", make_response(5, 9, 1, 2, 3, data=b"hello"), True),
    ("short", make_response(5, 9, 1, 2, 3)[:12], False),
    ("length long", make_response(5, 9, 1, 2, 3, length=15), False),
    ("length short", make_response(5, 9, 1, 2, 3, data=b"foo", length=15), False),
    ("bad data", make_response(5, 9, 1, 2, 3) + b"1", False),
    ("bad crc", make_response(5, 9, 1, 2, 3)[:-1] + b"1", False),
    ("bad data crc", make_response(5, 9, 1, 2, 3, data=b"foo")[:-1] + b"1", False),
)


@pytest.mark.parametrize(
    "packet, valid",
    [t[1:] for t in VALIDATE_RESPONSE_TESTS],
    ids=[t[0] for t in VALIDATE_RESPONSE_TESTS],
)
def test_validate_response(packet, valid):
    assert stentura._validate_response(packet) == valid


def test_read_data_simple():
    class MockPort:
        def read(self, count):
            if count != 5:
                raise Exception("Incorrect number read.")
            return b"12345"

    port = MockPort()
    buf = bytearray([0] * 20)
    count = stentura._read_data(port, threading.Event(), buf, 0, 5)
    assert count == 5
    assert buf == b"12345" + (b"\x00" * 15)

    # Test the offset parameter.
    count = stentura._read_data(port, threading.Event(), buf, 4, 5)
    assert buf == b"123412345" + (b"\x00" * 11)


def test_read_data_stop_set():
    class MockPort:
        def read(self, count):
            return b"0000"

    buf = bytearray()
    event = threading.Event()
    event.set()
    with pytest.raises(stentura._StopException):
        stentura._read_data(MockPort(), event, buf, 0, 4)


def test_read_data_timeout():
    class MockPort:
        def read(self, count):
            # When serial time out occurs read() returns
            # less characters as requested
            return "123"

    port = MockPort()
    buf = bytearray()
    with pytest.raises(stentura._TimeoutException):
        stentura._read_data(port, threading.Event(), buf, 0, 4)


def test_read_packet_simple():
    class MockPort:
        def __init__(self, packet):
            self._packet = packet

        def read(self, count):
            requested_bytes = self._packet[0:count]
            self._packet = self._packet[count:]
            return requested_bytes

    buf = bytearray(256)
    for packet in [
        make_response(1, 2, 3, 4, 5),
        make_response(1, 2, 3, 4, 5, b"hello"),
    ]:
        port = MockPort(packet)
        response = stentura._read_packet(port, threading.Event(), buf)
        assert response == packet


def test_read_packet_fail():
    class MockPort:
        def __init__(
            self,
            data_section_length=0,
            set1=False,
            set2=False,
            give_too_much_data=False,
            give_timeout=False,
        ):
            self._set1 = set1
            self._set2 = set2
            self._read1 = False
            self._read2 = False
            self.event = threading.Event()
            self._give_timeout = give_timeout
            self._data = [1, 0, data_section_length + 4, 0] + [0] * data_section_length
            if give_too_much_data:
                self._data.append(0)
            self._data = bytearray(self._data)

        def read(self, count):
            if not self._read1:
                self._read1 = True
                if self._set1:
                    self.event.set()
            elif not self._read2:
                self._read2 = True
                if self._set2:
                    self.event.set()
            else:
                raise Exception("Already read data.")
            if self._give_timeout and len(self._data) == count:
                # If read() returns less bytes what was requested,
                # it indicates a timeout.
                count -= 1
            requested_bytes = self._data[0:count]
            self._data = self._data[count:]
            return requested_bytes

    buf = bytearray()

    with pytest.raises(stentura._StopException):
        port = MockPort(set1=True)
        stentura._read_packet(port, port.event, buf)

    with pytest.raises(stentura._StopException):
        port = MockPort(data_section_length=30, set2=True)
        stentura._read_packet(port, port.event, buf)

    with pytest.raises(stentura._TimeoutException):
        port = MockPort(give_timeout=True)
        stentura._read_packet(port, port.event, buf)

    with pytest.raises(stentura._TimeoutException):
        port = MockPort(data_section_length=30, give_timeout=True)
        stentura._read_packet(port, port.event, buf)

    with pytest.raises(stentura._ProtocolViolationException):
        port = MockPort(give_too_much_data=True)
        stentura._read_packet(port, port.event, buf)


def test_write_to_port():
    class MockPort:
        def __init__(self, chunk):
            self._chunk = chunk
            self.data = b""

        def write(self, data):
            data = data[: self._chunk]
            self.data += data
            return len(data)

    data = bytearray(range(20))

    # All in one shot.
    port = MockPort(20)
    stentura._write_to_port(port, data)
    assert data == port.data

    # In parts.
    port = MockPort(5)
    stentura._write_to_port(port, data)
    assert data == port.data


def test_send_receive():
    event = threading.Event()
    buf, seq, action = bytearray(256), 5, stentura._OPEN
    request = stentura._make_request(bytearray(256), stentura._OPEN, seq)
    correct_response = make_response(seq, action)
    wrong_seq = make_response(seq - 1, action)
    wrong_action = make_response(seq, action + 1)
    bad_response = make_response(seq, action, data=b"foo", length=15)

    # Correct response first time.
    responses = [correct_response]
    port = MockPacketPort(responses)
    response = stentura._send_receive(port, event, request, buf)
    assert response == correct_response

    # Timeout once then correct response.
    responses = [b"", correct_response]
    port = MockPacketPort(responses)
    response = stentura._send_receive(port, event, request, buf)
    assert response == correct_response

    # Wrong sequence number then correct response.
    responses = [wrong_seq, correct_response]
    port = MockPacketPort(responses)
    response = stentura._send_receive(port, event, request, buf)
    assert response == correct_response

    # No correct responses. Also make sure max_retries is honored.
    max_tries = 6
    responses = [b""] * max_tries
    port = MockPacketPort(responses)
    with pytest.raises(stentura._ConnectionLostException):
        stentura._send_receive(port, event, request, buf, max_tries)
    assert max_tries == port.writes

    # Wrong action.
    responses = [wrong_action]
    port = MockPacketPort(responses)
    with pytest.raises(stentura._ProtocolViolationException):
        stentura._send_receive(port, event, request, buf)

    # Bad packet.
    responses = [bad_response]
    port = MockPacketPort(responses)
    with pytest.raises(stentura._ProtocolViolationException):
        stentura._send_receive(port, event, request, buf)

    # Stopped.
    responses = [""]
    event.set()
    port = MockPacketPort(responses)
    with pytest.raises(stentura._StopException):
        stentura._send_receive(port, event, request, buf)


def test_sequence_counter():
    seq = stentura._SequenceCounter()
    actual = [seq() for x in range(512)]
    expected = list(range(256)) * 2
    assert actual == expected

    seq = stentura._SequenceCounter(67)
    actual = [seq() for x in range(512)]
    expected = list(range(67, 256)) + list(range(256)) + list(range(67))
    assert actual == expected


def test_read():
    request_buf = bytearray(256)
    response_buf = bytearray(256)
    stroke_buf = bytearray(256)
    event = threading.Event()

    tests = ([0b11000001] * (3 * 512 + 28), [0b11010101] * 4, [0b11000010] * 8)

    for data in tests:
        data = bytearray(data)
        requests, responses = make_readc_packets(data)
        port = MockPacketPort(responses, requests)
        seq = stentura._SequenceCounter()
        block, byte = 0, 0
        block, byte, response = stentura._read(
            port, event, seq, request_buf, response_buf, stroke_buf, block, byte
        )
        assert data == bytes(response)
        assert block == len(data) // 512
        assert byte == len(data) % 512


def test_loop():
    class Event:
        def __init__(self, count, data, stop=False):
            self.count = count
            self.data = data
            self.stop = stop

        def __repr__(self):
            return "<{}, {}, {}>".format(self.count, self.data, self.stop)

    class MockPort:
        def __init__(self, events=[]):
            self._file = b""
            self._out = b""
            self._is_open = False
            self.event = threading.Event()
            self.count = 0
            self.events = [Event(*x) for x in sorted(events, key=lambda x: x[0])]

        def write(self, request):
            # Process the packet and put together a response.
            p = parse_request(request)
            if p["action"] == stentura._OPEN:
                self._out = make_response(p["seq"], p["action"])
                self._is_open = True
            elif p["action"] == stentura._READC:
                if not self._is_open:
                    raise Exception("no open")
                length, block, byte = p["p3"], p["p4"], p["p5"]
                seq = p["seq"]
                action = stentura._READC
                start = block * 512 + byte
                end = start + length
                data = self._file[start:end]
                self._out = make_response(seq, action, p1=len(data), data=data)
            while self.events and self.events[0].count <= self.count:
                event = self.events.pop(0)
                self.append(event.data)
                if event.stop:
                    self.event.set()
            self.count += 1
            return len(request)

        def read(self, count):
            requested_bytes = self._out[0:count]
            self._out = self._out[count:]
            return requested_bytes

        def append(self, data):
            self._file += data
            return self

        def flushInput(self):
            pass

        def flushOutput(self):
            pass

    data1 = bytearray([0b11001010] * 4 * 9)
    data1_trans = [["S-", "K-", "R-", "O-", "-F", "-P", "-T", "-D"]] * 9
    data2 = bytearray([0b11001011] * 4 * 30)

    tests = [
        # No inputs but nothing crashes either.
        (MockPort([(30, b"", True)]), []),
        # A few strokes.
        (MockPort([(23, data1), (43, b"", True)]), data1_trans),
        # Ignore data that's there before we started.
        (MockPort([(46, b"", True)]).append(data2), []),
        # Ignore data that was there and also parse some strokes.
        (MockPort([(25, data1), (36, b"", True)]).append(data2), data1_trans),
    ]

    for test in tests:
        read_data = []

        def callback(data):
            read_data.append(data)

        port = test[0]
        expected = test[1]

        ready_called = [False]

        def ready():
            ready_called[0] = True

        try:
            ready_called[0] = False
            stentura._loop(port, port.event, callback, ready, timeout=0)
        except stentura._StopException:
            pass
        assert read_data == expected
        assert ready_called[0]


# TODO: add a test on the machine itself with mocks



from collections import namedtuple
from pathlib import Path
from textwrap import dedent
from types import SimpleNamespace
import operator

from PySide6.QtCore import QModelIndex, QPersistentModelIndex, Qt

import pytest

from plover.config import DictionaryConfig
from plover.engine import ErroredDictionary
from plover.gui_qt.dictionaries_widget import DictionariesModel, DictionariesWidget
from plover.steno_dictionary import StenoDictionary, StenoDictionaryCollection
from plover.misc import expand_path

from plover_build_utils.testing import parametrize

from ..py37compat import mock


INVALID_EXCEPTION = Exception("loading error")

ICON_TO_CHAR = {
    "error": "!",
    "favorite": "★",
    "loading": "🗘",
    "normal": "⎕",
    "readonly": "🛇",
}
ICON_FROM_CHAR = {c: i for i, c in ICON_TO_CHAR.items()}

ENABLED_TO_CHAR = {
    False: "☐",
    True: "☑",
}
ENABLED_FROM_CHAR = {c: e for e, c in ENABLED_TO_CHAR.items()}

# TODO what does it mean that the MODEL_ROLES are now empty?
MODEL_ROLES = []
# before PySide6 refactoring:
# MODEL_ROLES = sorted([Qt.ItemDataRole.AccessibleTextRole, Qt.ItemDataRole.CheckStateRole,
#                      Qt.ItemDataRole.DecorationRole, Qt.ItemDataRole.DisplayRole,
#                      Qt.ItemDataRole.ToolTipRole])


def parse_state(state_str):
    state_str = dedent(state_str).strip()
    if not state_str:
        return
    for line in state_str.split("\n"):
        enabled, icon, path = line.split()
        yield ENABLED_FROM_CHAR[enabled], ICON_FROM_CHAR[icon], path


def config_dictionaries_from_state(state_str):
    return [
        DictionaryConfig(path, enabled)
        for enabled, icon, path in parse_state(state_str)
    ]


def steno_dictionaries_from_state(state_str, existing_dictionaries=None):
    new_dictionaries = []
    for enabled, icon, path in parse_state(state_str):
        if icon == "loading":
            continue
        path = expand_path(path)
        if existing_dictionaries is None:
            steno_dict = None
        else:
            steno_dict = existing_dictionaries.get(path)
        if steno_dict is None:
            if icon == "error" or path.endswith(".bad"):
                steno_dict = ErroredDictionary(path, INVALID_EXCEPTION)
            else:
                steno_dict = StenoDictionary()
                steno_dict.path = path
                steno_dict.readonly = (
                    icon == "readonly"
                    or path.endswith(".ro")
                    or path.startswith("asset:")
                )
            steno_dict.enabled = enabled
        new_dictionaries.append(steno_dict)
    return new_dictionaries


class ModelTest(
    namedtuple(
        "ModelTest",
        """
                           config dictionaries engine
                           model signals connections
                           initial_state
                           """,
    )
):
    def configure(self, **kwargs):
        self.connections["config_changed"](kwargs)

    def configure_dictionaries(self, state):
        self.configure(dictionaries=config_dictionaries_from_state(state))

    def load_dictionaries(self, state):
        self.dictionaries.set_dicts(
            steno_dictionaries_from_state(state, self.dictionaries)
        )
        self.connections["dictionaries_loaded"](self.dictionaries)
        loaded = [
            row
            for row, (enabled, icon, path) in enumerate(parse_state(state))
            if icon != "loading"
        ]

    def check(
        self,
        expected,
        config_change=None,
        data_change=None,
        layout_change=False,
        undo_change=None,
    ):
        __tracebackhide__ = operator.methodcaller("errisinstance", AssertionError)
        expected = dedent(expected).strip()
        if expected:
            expected_config = expected
            expected_state = expected.split("\n")
        else:
            expected_config = ""
            expected_state = []
        actual_state = []
        for row in range(self.model.rowCount()):
            index = self.model.index(row)
            check_state = index.data(Qt.ItemDataRole.CheckStateRole)
            is_checked = check_state == Qt.CheckState.Checked
            icon = index.data(Qt.ItemDataRole.DecorationRole)
            path = index.data(Qt.ItemDataRole.DisplayRole)
            actual_state.append(
                "%s %s %s"
                % (
                    ENABLED_TO_CHAR.get(is_checked, "?"),
                    ICON_TO_CHAR.get(icon, "?"),
                    path,
                )
            )
        assert actual_state == expected_state
        assert not self.engine.mock_calls, "unexpected engine call"
        if config_change == "reload":
            assert self.config.mock_calls == [mock.call({})]
            self.config.reset_mock()
        elif config_change == "update":
            config_update = {
                "dictionaries": config_dictionaries_from_state(expected_config),
            }
            assert self.config.mock_calls == [mock.call(config_update)]
            self.config.reset_mock()
        else:
            assert not self.config.mock_calls, "unexpected config call"
        signal_calls = self.signals.mock_calls[:]
        if undo_change is not None:
            call = signal_calls.pop(0)
            assert call == mock.call.has_undo_changed(undo_change)
        if data_change is not None:
            for row in data_change:
                index = self.model.index(row)
                call = signal_calls.pop(0)
                call.args[2].sort()
                assert call == mock.call.dataChanged(index, index, MODEL_ROLES)
        if layout_change:
            assert signal_calls[0:2] == [
                mock.call.layoutAboutToBeChanged(),
                mock.call.layoutChanged(),
            ]
            del signal_calls[0:2]
        assert not signal_calls
        self.signals.reset_mock()

    def reset_mocks(self):
        self.config.reset_mock()
        self.engine.reset_mock()
        self.signals.reset_mock()


@pytest.fixture
def model_test(monkeypatch, request):
    state = request.function.__doc__
    # Patch configuration directory.
    current_dir = Path(".").resolve()
    monkeypatch.setattr("plover.misc.CONFIG_DIR", str(current_dir))
    monkeypatch.setattr(
        "plover.gui_qt.dictionaries_widget.CONFIG_DIR", str(current_dir)
    )
    # Disable i18n support.
    monkeypatch.setattr("plover.gui_qt.dictionaries_widget._", lambda s: s)
    # Fake config.
    config = mock.PropertyMock()
    config.return_value = {}
    # Dictionaries.
    dictionaries = StenoDictionaryCollection()
    # Fake engine.
    engine = mock.MagicMock(
        spec="""
                            __enter__ __exit__
                            config signal_connect
                            """.split()
    )
    engine.__enter__.return_value = engine
    type(engine).config = config
    signals = mock.MagicMock()
    config.return_value = {
        "dictionaries": config_dictionaries_from_state(state) if state else [],
        "classic_dictionaries_display_order": False,
    }
    # Setup model.
    model = DictionariesModel(engine, {name: name for name in ICON_TO_CHAR}, max_undo=5)
    for slot in """
    dataChanged
    layoutAboutToBeChanged
    layoutChanged
    has_undo_changed
    """.split():
        getattr(model, slot).connect(getattr(signals, slot))
    connections = dict(call.args for call in engine.signal_connect.mock_calls)
    assert connections.keys() == {
        "config_changed",
        "dictionaries_loaded",
        "dictionary_state_changed",
    }
    config.reset_mock()
    engine.reset_mock()
    # Test helper.
    test = ModelTest(config, dictionaries, engine, model, signals, connections, state)
    if state and any(icon != "loading" for enabled, icon, path in parse_state(state)):
        test.load_dictionaries(state)
        test.reset_mocks()
    return test


def test_model_accessible_text_1(model_test):
    """
    ☑ 🗘 read-only.ro
    ☑ 🗘 user.json
    ☑ 🗘 invalid.bad
    ☑ 🗘 commands.json
    ☐ 🗘 asset:plover:assets/main.json
    """
    for n, expected in enumerate(
        (
            "read-only.ro, loading",
            "user.json, loading",
            "invalid.bad, loading",
            "commands.json, loading",
            "asset:plover:assets/main.json, disabled, loading",
        )
    ):
        assert (
            model_test.model.index(n).data(Qt.ItemDataRole.AccessibleTextRole)
            == expected
        )


def test_model_accessible_text_2(model_test):
    """
    ☑ 🛇 read-only.ro
    ☑ ★ user.json
    ☑ ⎕ commands.json
    ☐ 🛇 asset:plover:assets/main.json
    """
    for n, expected in enumerate(
        (
            "read-only.ro, read-only",
            "user.json, favorite",
            "commands.json",
            "asset:plover:assets/main.json, disabled, read-only",
        )
    ):
        assert (
            model_test.model.index(n).data(Qt.ItemDataRole.AccessibleTextRole)
            == expected
        )


def test_model_accessible_text_3(model_test):
    """
    ☑ ! invalid.bad
    """
    expected = "invalid.bad, errored: %s." % INVALID_EXCEPTION
    assert (
        model_test.model.index(0).data(Qt.ItemDataRole.AccessibleTextRole) == expected
    )


def test_model_accessible_text_4(model_test):
    """
    ☐ ! invalid.bad
    """
    expected = "invalid.bad, disabled, errored: %s." % INVALID_EXCEPTION
    assert (
        model_test.model.index(0).data(Qt.ItemDataRole.AccessibleTextRole) == expected
    )


def test_model_add_existing(model_test):
    """
    ☑ ★ user.json
    ☑ ⎕ commands.json
    ☐ ⎕ main.json
    """
    model_test.model.add([expand_path("main.json")])
    model_test.check(model_test.initial_state, config_change="reload")


def test_model_add_new_1(model_test):
    """
    ☑ ★ user.json
    ☐ ⎕ commands.json
    ☑ 🗘 main.json
    """
    model_test.model.add([expand_path("read-only.ro")])
    model_test.check(
        """
        ☑ 🗘 read-only.ro
        ☑ ★ user.json
        ☐ ⎕ commands.json
        ☑ 🗘 main.json
        """,
        config_change="update",
        layout_change=True,
        undo_change=True,
    )


def test_model_add_new_2(model_test):
    """
    ☑ ★ user.json
    ☐ ⎕ commands.json
    ☑ 🗘 main.json
    """
    model_test.model.add(["duplicated.json", "unique.json", "duplicated.json"])
    model_test.check(
        """
        ☑ 🗘 duplicated.json
        ☑ 🗘 unique.json
        ☑ ★ user.json
        ☐ ⎕ commands.json
        ☑ 🗘 main.json
        """,
        config_change="update",
        layout_change=True,
        undo_change=True,
    )


def test_model_add_nothing(model_test):
    """
    ☑ ★ user.json
    ☑ ⎕ commands.json
    ☑ ⎕ main.json
    """
    model_test.model.add([])
    model_test.check(model_test.initial_state)


def test_model_config_update(model_test):
    """
    ☐ ⎕ user.json
    ☑ ★ commands.json
    ☑ 🛇 read-only.ro
    ☑ ! invalid.bad
    ☐ 🛇 asset:plover:assets/main.json
    """
    state = """
    ☑ ★ user.json
    ☐ ⎕ commands.json
    ☑ 🗘 main.json
    """
    model_test.configure_dictionaries(state)
    model_test.check(state, layout_change=True)
    state = """
    ☑ ★ user.json
    ☐ ⎕ commands.json
    ☑ ⎕ main.json
    """
    model_test.load_dictionaries(state)
    model_test.check(state, data_change=[2])


def test_model_insert_1(model_test):
    """
    ☐ ⎕ user.json
    ☑ ★ commands.json
    ☑ 🛇 read-only.ro
    ☑ ! invalid.bad
    ☐ 🛇 asset:plover:assets/main.json
    """
    model_test.model.insert(
        model_test.model.index(2), ["main.json", "commands.json", "read-only.ro"]
    )
    model_test.check(
        """
        ☐ ⎕ user.json
        ☑ 🗘 main.json
        ☑ ★ commands.json
        ☑ 🛇 read-only.ro
        ☑ ! invalid.bad
        ☐ 🛇 asset:plover:assets/main.json
        """,
        config_change="update",
        layout_change=True,
        undo_change=True,
    )


def test_model_insert_2(model_test):
    """
    ☐ ⎕ user.json
    ☑ 🗘 main.json
    ☑ ★ commands.json
    ☑ 🛇 read-only.ro
    ☑ ! invalid.bad
    ☐ 🛇 asset:plover:assets/main.json
    """
    model_test.model.insert(
        QModelIndex(), ["commands.json", "user.json", "commands.json"]
    )
    model_test.check(
        """
        ☑ 🗘 main.json
        ☑ 🛇 read-only.ro
        ☑ ! invalid.bad
        ☐ 🛇 asset:plover:assets/main.json
        ☑ ★ commands.json
        ☐ ⎕ user.json
        """,
        config_change="update",
        layout_change=True,
        undo_change=True,
    )


def test_model_insert_3(model_test):
    """
    ☑ ★ user.json
    ☑ ⎕ commands.json
    ☑ ⎕ main.json
    """
    model_test.model.insert(QModelIndex(), [])
    model_test.check(model_test.initial_state)


def test_model_display_order(model_test):
    """
    ☑ 🛇 read-only.ro
    ☑ ★ user.json
    ☑ ⎕ commands.json
    ☐ 🛇 asset:plover:assets/main.json
    """
    state = model_test.initial_state
    # Flip display order.
    model_test.configure(classic_dictionaries_display_order=True)
    model_test.check("\n".join(reversed(state.split("\n"))), layout_change=True)
    # Reset display order to default.
    model_test.configure(classic_dictionaries_display_order=False)
    model_test.check(state, layout_change=True)


def test_model_favorite(model_test):
    """
    ☑ 🛇 read-only.ro
    ☑ ★ user.json
    ☑ ! invalid.bad
    ☑ ⎕ commands.json
    ☐ 🛇 asset:plover:assets/main.json
    """
    # New favorite.
    model_test.model.setData(
        model_test.model.index(1),
        Qt.CheckState.Unchecked.value,
        Qt.ItemDataRole.CheckStateRole.value,
    )
    model_test.check(
        """
        ☑ 🛇 read-only.ro
        ☐ ⎕ user.json
        ☑ ! invalid.bad
        ☑ ★ commands.json
        ☐ 🛇 asset:plover:assets/main.json
        """,
        config_change="update",
        data_change=[1, 3],
        undo_change=True,
    )
    # No favorite.
    model_test.model.setData(
        model_test.model.index(3),
        Qt.CheckState.Unchecked.value,
        Qt.ItemDataRole.CheckStateRole.value,
    )
    model_test.check(
        """
        ☑ 🛇 read-only.ro
        ☐ ⎕ user.json
        ☑ ! invalid.bad
        ☐ ⎕ commands.json
        ☐ 🛇 asset:plover:assets/main.json
        """,
        config_change="update",
        data_change=[3],
    )


def test_model_initial_setup(model_test):
    """
    ☑ 🗘 read-only.ro
    ☑ 🗘 user.json
    ☑ 🗘 invalid.bad
    ☑ 🗘 commands.json
    ☐ 🗘 asset:plover:assets/main.json
    """
    state = model_test.initial_state
    # Initial state.
    model_test.check(state)
    # First config notification: no-op.
    model_test.configure(**model_test.config.return_value)
    model_test.check(state)
    # After loading dictionaries.
    state = """
    ☑ 🛇 read-only.ro
    ☑ ★ user.json
    ☑ ! invalid.bad
    ☑ ⎕ commands.json
    ☐ 🛇 asset:plover:assets/main.json
    """
    model_test.load_dictionaries(state)
    model_test.check(state, data_change=range(5))


def test_model_iter_loaded(model_test):
    """
    ☑ 🗘 magnum.json
    ☑ ★ user.json
    ☐ ⎕ commands.json
    ☑ 🗘 main.json
    """
    model_test.check(model_test.initial_state)
    index_list = [model_test.model.index(n) for n in range(4)]
    index_list.append(QModelIndex())
    assert (
        list(model_test.model.iter_loaded(index_list)) == model_test.dictionaries.dicts
    )
    assert (
        list(model_test.model.iter_loaded(reversed(index_list)))
        == model_test.dictionaries.dicts
    )
    model_test.configure(classic_dictionaries_display_order=False)
    assert (
        list(model_test.model.iter_loaded(index_list)) == model_test.dictionaries.dicts
    )
    assert (
        list(model_test.model.iter_loaded(reversed(index_list)))
        == model_test.dictionaries.dicts
    )


def test_model_move_dictionaries(model_test):
    """
    ☑ 🛇 read-only.ro
    ☐ ⎕ commands.json
    ☑ ★ user.json
    ☑ ⎕ main.json
    """
    model_test.check(model_test.initial_state)
    model_test.model.move(
        QModelIndex(), [model_test.model.index(0), model_test.model.index(2)]
    )
    model_test.check(
        """
        ☐ ⎕ commands.json
        ☑ ★ main.json
        ☑ 🛇 read-only.ro
        ☑ ⎕ user.json
        """,
        config_change="update",
        layout_change=True,
        undo_change=True,
    )


def test_model_move_down(model_test):
    """
    ☐ ⎕ commands.json
    ☑ ★ main.json
    ☑ 🛇 read-only.ro
    ☑ ⎕ user.json
    """
    model_test.model.move_down([model_test.model.index(n) for n in [1]])
    model_test.check(
        """
        ☐ ⎕ commands.json
        ☑ 🛇 read-only.ro
        ☑ ★ main.json
        ☑ ⎕ user.json
        """,
        config_change="update",
        layout_change=True,
        undo_change=True,
    )
    model_test.model.move_down([model_test.model.index(n) for n in [0, 2]])
    model_test.check(
        """
        ☑ 🛇 read-only.ro
        ☐ ⎕ commands.json
        ☑ ★ user.json
        ☑ ⎕ main.json
        """,
        config_change="update",
        layout_change=True,
    )
    model_test.model.move_down([model_test.model.index(n) for n in [1]])
    model_test.check(
        """
        ☑ 🛇 read-only.ro
        ☑ ★ user.json
        ☐ ⎕ commands.json
        ☑ ⎕ main.json
        """,
        config_change="update",
        layout_change=True,
    )


def test_model_move_down_nothing(model_test):
    """
    ☑ ★ user.json
    ☑ ⎕ commands.json
    ☑ ⎕ main.json
    """
    model_test.model.move_down([QModelIndex()])
    model_test.check(model_test.initial_state)


def test_model_move_up(model_test):
    """
    ☑ 🛇 read-only.ro
    ☑ ⎕ user.json
    ☐ ⎕ commands.json
    ☑ ★ main.json
    """
    model_test.model.move_up([model_test.model.index(n) for n in [2, 3]])
    model_test.check(
        """
        ☑ 🛇 read-only.ro
        ☐ ⎕ commands.json
        ☑ ★ main.json
        ☑ ⎕ user.json
        """,
        config_change="update",
        layout_change=True,
        undo_change=True,
    )
    model_test.model.move_up([model_test.model.index(n) for n in [1, 2]])
    model_test.check(
        """
        ☐ ⎕ commands.json
        ☑ ★ main.json
        ☑ 🛇 read-only.ro
        ☑ ⎕ user.json
        """,
        config_change="update",
        layout_change=True,
    )


def test_model_move_up_nothing(model_test):
    """
    ☑ ★ user.json
    ☑ ⎕ commands.json
    ☑ ⎕ main.json
    """
    model_test.model.move_up([QModelIndex()])
    model_test.check(model_test.initial_state)


def test_model_persistent_index(model_test):
    """
    ☑ 🛇 read-only.ro
    ☑ ★ user.json
    ☑ ⎕ commands.json
    ☐ 🛇 asset:plover:assets/main.json
    """
    persistent_index = QPersistentModelIndex(model_test.model.index(1))
    assert persistent_index.row() == 1
    assert (
        persistent_index.data(Qt.ItemDataRole.CheckStateRole) == Qt.CheckState.Checked
    )
    assert persistent_index.data(Qt.ItemDataRole.DecorationRole) == "favorite"
    assert persistent_index.data(Qt.ItemDataRole.DisplayRole) == "user.json"
    model_test.configure(classic_dictionaries_display_order=True)
    assert persistent_index.row() == 2
    assert (
        persistent_index.data(Qt.ItemDataRole.CheckStateRole) == Qt.CheckState.Checked
    )
    assert persistent_index.data(Qt.ItemDataRole.DecorationRole) == "favorite"
    assert persistent_index.data(Qt.ItemDataRole.DisplayRole) == "user.json"
    model_test.model.setData(
        persistent_index,
        Qt.CheckState.Unchecked.value,
        Qt.ItemDataRole.CheckStateRole.value,
    )
    assert persistent_index.row() == 2
    assert (
        persistent_index.data(Qt.ItemDataRole.CheckStateRole) == Qt.CheckState.Unchecked
    )
    assert persistent_index.data(Qt.ItemDataRole.DecorationRole) == "normal"
    assert persistent_index.data(Qt.ItemDataRole.DisplayRole) == "user.json"


def test_model_qtmodeltester(model_test, qtmodeltester):
    """
    ☑ 🛇 read-only.ro
    ☑ ★ user.json
    ☑ ⎕ commands.json
    ☐ 🛇 asset:plover:assets/main.json
    """
    qtmodeltester.check(model_test.model)


def test_model_remove(model_test):
    """
    ☐ ⎕ commands.json
    ☑ 🛇 read-only.ro
    ☑ ★ main.json
    ☑ ⎕ user.json
    """
    model_test.model.remove([model_test.model.index(n) for n in [0, 3]])
    model_test.check(
        """
        ☑ 🛇 read-only.ro
        ☑ ★ main.json
        """,
        config_change="update",
        layout_change=True,
        undo_change=True,
    )
    model_test.model.remove([model_test.model.index(n) for n in [0, 1]])
    model_test.check("", config_change="update", layout_change=True)


def test_model_remove_nothing_1(model_test):
    """
    ☐ ⎕ commands.json
    ☑ 🛇 read-only.ro
    ☑ ★ main.json
    ☑ ⎕ user.json
    """
    model_test.model.remove([])
    model_test.check(model_test.initial_state)


def test_model_remove_nothing_2(model_test):
    """
    ☐ ⎕ commands.json
    ☑ 🛇 read-only.ro
    ☑ ★ main.json
    ☑ ⎕ user.json
    """
    model_test.model.remove([QModelIndex()])
    model_test.check(model_test.initial_state)


def test_model_set_checked(model_test):
    on_state = "☑ 🗘 user.json"
    off_state = "☐ 🗘 user.json"
    model_test.model.add(["user.json"])
    model_test.check(
        on_state, config_change="update", layout_change=True, undo_change=True
    )
    first_index = model_test.model.index(0)
    for index, value, ret, state in (
        # Invalid index.
        (QModelIndex(), Qt.CheckState.Unchecked.value, False, on_state),
        # Invalid values.
        (first_index, "pouet", False, on_state),
        (first_index, Qt.CheckState.PartiallyChecked.value, False, on_state),
        # Already checked.
        (first_index, Qt.CheckState.Checked.value, False, on_state),
        # Uncheck.
        (first_index, Qt.CheckState.Unchecked.value, True, off_state),
        # Recheck.
        (first_index, Qt.CheckState.Checked.value, True, on_state),
    ):
        assert (
            model_test.model.setData(index, value, Qt.ItemDataRole.CheckStateRole.value)
            == ret
        )
        model_test.check(
            state,
            config_change="update" if ret else None,
            data_change=[index.row()] if ret else None,
        )


def test_model_unrelated_config_change(model_test):
    """
    ☑ ★ user.json
    ☑ ⎕ commands.json
    ☑ ⎕ main.json
    """
    model_test.configure(start_minimized=True)
    model_test.check(model_test.initial_state)


def test_model_undo_1(model_test):
    """
    ☐ 🗘 1.json
    ☐ 🗘 2.json
    ☐ 🗘 3.json
    ☐ 🗘 4.json
    ☐ 🗘 5.json
    ☐ 🗘 6.json
    """
    # Check max undo size.
    state = dedent(model_test.initial_state).strip()
    state_stack = []
    for n in range(6):
        state_stack.append(state)
        state = state.split("\n")
        state[n] = "☑" + state[n][1:]
        state = "\n".join(state)
        model_test.model.setData(
            model_test.model.index(n),
            Qt.CheckState.Checked.value,
            Qt.ItemDataRole.CheckStateRole.value,
        )
        model_test.check(
            state,
            config_change="update",
            data_change=[n],
            undo_change=(True if n == 0 else None),
        )
    for n in range(5):
        model_test.model.undo()
        model_test.check(
            state_stack.pop(),
            config_change="update",
            layout_change=True,
            undo_change=(False if n == 4 else None),
        )


def test_model_undo_2(model_test):
    """
    ☑ ★ user.json
    ☑ ⎕ commands.json
    ☑ ⎕ main.json
    """
    # Changing display order as no impact on the undo stack.
    model_test.configure(classic_dictionaries_display_order=True)
    model_test.check(
        """
        ☑ ⎕ main.json
        ☑ ⎕ commands.json
        ☑ ★ user.json
        """,
        layout_change=True,
    )


class WidgetTest(
    namedtuple(
        "WidgetTest",
        """
                            registry
                            bot widget
                            file_dialog
                            create_dictionary
                            model_test
                            """,
    )
):
    def select(self, selection):
        sm = self.widget.view.selectionModel()
        for row in selection:
            sm.select(self.model.index(row), sm.SelectionFlag.Select)

    def unselect(self, selection):
        sm = self.widget.view.selectionModel()
        for row in selection:
            sm.select(self.model.index(row), sm.SelectionFlag.Deselect)

    def __getattr__(self, name):
        return getattr(self.model_test, name)


@pytest.fixture
def widget_test(model_test, monkeypatch, qtbot):
    # Fake registry.
    def list_plugins(plugin_type):
        assert plugin_type == "dictionary"
        for name, readonly in (
            ("bad", False),
            ("json", False),
            ("ro", True),
        ):
            obj = SimpleNamespace(readonly=readonly)
            yield SimpleNamespace(name=name, obj=obj)

    registry = mock.MagicMock(spec=["list_plugins"])
    registry.list_plugins.side_effect = list_plugins
    monkeypatch.setattr("plover.gui_qt.dictionaries_widget.registry", registry)
    # Fake file dialog.
    file_dialog = mock.MagicMock(
        spec="""
                                 getOpenFileNames
                                 getSaveFileName
                                 """.split()
    )
    monkeypatch.setattr("plover.gui_qt.dictionaries_widget.QFileDialog", file_dialog)

    # Fake `create_dictionary`.
    def create_dictionary(filename, threaded_save=True):
        pass

    steno_dict = mock.create_autospec(StenoDictionary)
    create_dictionary = mock.create_autospec(create_dictionary, return_value=steno_dict)
    monkeypatch.setattr(
        "plover.gui_qt.dictionaries_widget.create_dictionary", create_dictionary
    )
    # Patch `DictionariesModel` constructor to use our own instance.
    monkeypatch.setattr(
        "plover.gui_qt.dictionaries_widget.DictionariesModel",
        lambda engine, icons: model_test.model,
    )
    widget = DictionariesWidget()
    widget.setup(model_test.engine)
    qtbot.addWidget(widget)
    test = WidgetTest(
        registry, qtbot, widget, file_dialog, create_dictionary, model_test
    )
    return test


@parametrize(
    (
        # No selection.
        lambda: (
            (),
            """
     AddDictionaries
     AddTranslation
     """,
        ),
        # No loaded dictionary selected.
        lambda: (
            [1, 4],
            """
     AddDictionaries
     AddTranslation
     MoveDictionariesDown
     MoveDictionariesUp
     RemoveDictionaries
     """,
        ),
        # At least one loaded dictionary selected.
        lambda: (
            [0, 2],
            """
     AddDictionaries
     AddTranslation
     EditDictionaries
     MoveDictionariesDown
     MoveDictionariesUp
     RemoveDictionaries
     SaveDictionaries
     """,
        ),
        lambda: (
            [1, 3],
            """
     AddDictionaries
     AddTranslation
     EditDictionaries
     MoveDictionariesDown
     MoveDictionariesUp
     RemoveDictionaries
     SaveDictionaries
     """,
        ),
    )
)
def test_widget_selection(widget_test, selection, enabled_actions):
    """
    ☑ ★ favorite.json
    ☑ 🗘 loading.json
    ☑ ⎕ normal.json
    ☑ 🛇 read-only.ro
    ☑ ! invalid.bad
    """
    widget_test.select(selection)
    for action_name in """
    AddDictionaries
    AddTranslation
    EditDictionaries
    MoveDictionariesDown
    MoveDictionariesUp
    RemoveDictionaries
    SaveDictionaries
    Undo
    """.split():
        action = getattr(widget_test.widget, "action_" + action_name)
        enabled = action.isEnabled()
        msg = "%s is %s" % (action_name, "enabled" if enabled else "disabled")
        assert enabled == (action_name in enabled_actions), msg


FILE_PICKER_SAVE_FILTER = "Dictionaries (*.bad *.json);; BAD dictionaries (*.bad);; JSON dictionaries (*.json)"


def test_widget_save_copy_1(widget_test):
    """
    ☑ ★ favorite.json
    ☑ 🗘 loading.json
    ☑ ⎕ normal.json
    ☑ 🛇 read-only.ro
    ☑ ! invalid.bad
    """
    # Setup.
    copy_names = (
        expand_path("favorite_copy.json"),
        "",
        expand_path("read-only_copy.json"),
    )
    widget_test.file_dialog.getSaveFileName.side_effect = [
        [name] for name in copy_names
    ]
    steno_dict_copies = (
        mock.create_autospec(StenoDictionary),
        mock.create_autospec(StenoDictionary),
    )
    widget_test.create_dictionary.side_effect = steno_dict_copies
    # Execution.
    widget_test.select(range(5))
    widget_test.widget.action_CopyDictionaries.trigger()
    # Check.
    assert widget_test.file_dialog.mock_calls == [
        mock.call.getSaveFileName(
            parent=widget_test.widget,
            caption="Save a copy of %s as..." % name,
            dir=expand_path("%s - Copy.json" % Path(name).stem),
            filter=FILE_PICKER_SAVE_FILTER,
        )
        for name in ["favorite.json", "normal.json", "read-only.ro"]
    ]
    assert widget_test.create_dictionary.mock_calls == [
        mock.call(name, threaded_save=False) for name in copy_names if name
    ]
    assert steno_dict_copies[0].mock_calls == [
        mock.call.update(widget_test.dictionaries.dicts[0]),
        mock.call.save(),
    ]
    assert steno_dict_copies[1].mock_calls == [
        mock.call.update(widget_test.dictionaries.dicts[2]),
        mock.call.save(),
    ]


def test_widget_save_merge_1(widget_test):
    """
    ☑ ★ favorite.json
    ☑ 🗘 loading.json
    ☑ ⎕ normal.json
    ☑ 🛇 read-only.ro
    ☑ ! invalid.bad
    """
    # Setup.
    merge_name = "favorite + normal + read-only"
    widget_test.file_dialog.getSaveFileName.return_value = [expand_path("merge.json")]
    # Execution.
    widget_test.select(range(5))
    widget_test.widget.action_MergeDictionaries.trigger()
    # Check.
    assert widget_test.file_dialog.mock_calls == [
        mock.call.getSaveFileName(
            parent=widget_test.widget,
            caption="Merge %s as..." % merge_name,
            dir=expand_path(merge_name + ".json"),
            filter=FILE_PICKER_SAVE_FILTER,
        )
    ]
    assert widget_test.create_dictionary.mock_calls == [
        mock.call(expand_path("merge.json"), threaded_save=False)
    ]
    steno_dict = widget_test.create_dictionary.return_value
    assert steno_dict.mock_calls == [
        mock.call.update(widget_test.dictionaries.dicts[2]),
        mock.call.update(widget_test.dictionaries.dicts[1]),
        mock.call.update(widget_test.dictionaries.dicts[0]),
        mock.call.save(),
    ]


def test_widget_save_merge_2(widget_test):
    """
    ☑ ★ favorite.json
    ☑ 🗘 loading.json
    ☑ ⎕ normal.json
    ☑ 🛇 read-only.ro
    ☑ ! invalid.bad
    """
    # Setup.
    merge_name = "favorite + normal"
    widget_test.file_dialog.getSaveFileName.return_value = [""]
    # Execution.
    widget_test.select([0, 2])
    widget_test.widget.action_MergeDictionaries.trigger()
    # Check.
    assert widget_test.file_dialog.mock_calls == [
        mock.call.getSaveFileName(
            parent=widget_test.widget,
            caption="Merge %s as..." % merge_name,
            dir=expand_path(merge_name + ".json"),
            filter=FILE_PICKER_SAVE_FILTER,
        )
    ]
    assert widget_test.create_dictionary.mock_calls == []
